"""
This is a module for a more accurate estimation of TES parameters than what
currently exists in the IV and Bias Step Modules.

It contains the follwing procedures:
  - recompute_ivpars: corrects errors in IV analysis
  - run_correction: Takes IV and bias step data to compute corrected TES params
    based on the RP fitting method described here:
        https://simonsobs.atlassian.net/wiki/spaces/~5570586d07625a6be74c8780e4b96f6156f5e6/blog/2024/02/02/286228683/Nonlinear+TES+model+using+RP+curve

Authors: Remy Gerras, Satoru Takakura, Jack Lashner
"""

import numpy as np
from scipy.interpolate import interp1d
from scipy.optimize import minimize
from tqdm.auto import trange
import multiprocessing
from concurrent.futures import ProcessPoolExecutor, as_completed, Future
from typing import Optional, List, Tuple, TypeVar
from dataclasses import dataclass
import traceback
from copy import deepcopy
from numpy.typing import NDArray

from sodetlib.operations.iv import IVAnalysis
from sodetlib.operations.bias_steps import BiasStepAnalysis

D = TypeVar("D", float, np.ndarray)


def model_logalpha(r: D, logp0=0.0, p1=7.0, p2=1.0, logp3=0.0) -> D:
    """
    Fits Alpha' parameter, or alpha / GT. P0 has units of [uW]^-1.
    """
    return logp0 + p1 * r - p2 * r ** np.exp(logp3) * np.log(1 - r)


@dataclass
class AnalysisCfg:
    """
    Class for configuring behavior of the parameter estimation functions
    """

    default_nprocs: int = 10
    "Default number of processes to use when running correction in parallel."

    raise_exceptions: bool = False
    "If true, will raise exceptions thrown in the run_correction function."

    psat_level: float = 0.9
    "rfrac for which psat is defined."

    rpfit_min_rfrac: float = 0.05
    "Min rfrac used for RP fits and estimation."

    rpfit_max_rfrac: float = 0.95
    "Max rfrac used for RP fits and estiamtion."

    rpfit_intg_nsamps: int = 500
    "Number of samples to use in the RP fit and integral."

    sc_rfrac_thresh: float = 0.02
    "Bias step rfrac threshold below which TES should be considered "
    "superconducting, and the correction skipped."

    show_pb: bool = True
    "If true, will show a pb when running on all channels together"

    def __post_init__(self):
        if self.rpfit_min_rfrac < 0.0001:
            raise ValueError("rp_fit_min_rfrac must be larger than 0.0001")
        if self.rpfit_max_rfrac > 0.9999:
            raise ValueError("rp_fit_max_rfrac must be smaller than 0.9999")


@dataclass
class RpFitChanData:
    """
    Channel data needed to perform RP correction, pulled from bias step and IV
    calibration functions.

    Args
    -------
    bg: int
        Bias Line
    Rsh: float
        Shunt Resistance [Ohms]
    band: int
        Readout band
    channel :int
        Readout channel
    iv_idx_sc: int
        Index in IV data for SC transition
    iv_resp: np.ndarray
        TES current response relative from IV.
    iv_R: np.ndarray
        TES resistance [Ohms] through IV.
    iv_p_tes: np.ndarray
        TES Bias power [pW] through IV.
    iv_Rn: float
        TES normal resistance, measured from IV.
    iv_ibias: np.ndarray
        TES bias current [uA] throughout IV.
    iv_si: np.ndarray
        TES responsivity [1/uV] throughout IV.
    bs_meas_dIrat: float
        TES measured dIdrat from bias steps.
    bs_Rtes: float
        TES resistance [Ohm] measured from bias steps.
    bs_Ibias: float
        TES bias current [uA] measured from bias steps.
    bs_Si: float
        TES responsivity [1/uV] measured from bias steps.
    """

    bg: int
    Rsh: float

    band: int
    channel: int

    iv_idx_sc: int
    iv_resp: np.ndarray
    iv_R: np.ndarray  # Ohm
    iv_p_tes: np.ndarray  # pW
    iv_Rn: float
    iv_ibias: np.ndarray  # uA
    iv_si: np.ndarray  # 1/uV

    bs_meas_dIrat: float
    bs_Rtes: float = np.nan  # Ohm
    bs_Ibias: float = np.nan  # uA
    bs_Si: float = np.nan  # 1/uV

    @classmethod
    def from_data(
        cls, iva: IVAnalysis, bsa: BiasStepAnalysis, band: int, channel: int
    ) -> "RpFitChanData":
        """Create class from IV and BSA dicts"""
        iv_idx = np.where((iva.channels == channel) & (iva.bands == band))[0]
        if len(iv_idx) == 0:
            raise ValueError(
                f"Could not find band={band} channel={channel} in IVAnalysis"
            )
        iv_idx = iv_idx[0]

        bs_idx = np.where((bsa.channels == channel) & (bsa.bands == band))[0]
        if len(bs_idx) == 0:
            raise ValueError(
                f"Could not find band={band} channel={channel} in Bias Steps"
            )
        bs_idx = bs_idx[0]

        bg = iva.bgmap[iv_idx]

        return cls(
            bg=bg,
            band=band,
            channel=channel,
            Rsh=iva.meta["R_sh"],
            iv_idx_sc=iva.idxs[iv_idx, 0],
            iv_resp=iva.resp[iv_idx],
            iv_R=iva.R[iv_idx],
            iv_Rn=iva.R_n[iv_idx],
            iv_p_tes=iva.p_tes[iv_idx] * 1e12,
            iv_si=iva.si[iv_idx] * 1e-6,
            iv_ibias=iva.i_bias * 1e6,
            bs_meas_dIrat=bsa.dItes[bs_idx] / bsa.dIbias[bg],
            bs_Rtes=bsa.R0[bs_idx],
            bs_Ibias=bsa.Ibias[bg] * 1e6,
            bs_Si=bsa.Si[bs_idx] * 1e-6,
        )


@dataclass
class CorrectedParams:
    "Log(alpha') fit popts"

    logalpha_popts: np.ndarray
    "Log alpha fit parameters"
    pbias_model_offset: float
    "delta Popt between IV and bias steps"
    delta_Popt: float
    "TES Current  [uA]"
    corrected_I0: float
    "TES resistance [Ohm]"
    corrected_R0: float
    "Bias power [aW]"
    corrected_Pj: float
    "Responsivity [1/uV]"
    corrected_Si: float
    "Loopgain at bias current"
    loopgain: float


@dataclass
class CorrectionResults:
    """
    Results from the parameter correction procedure
    """

    chdata: RpFitChanData
    "Channel data used for correction"

    success: bool = False
    "If correction finished successfully"

    traceback: Optional[str] = None
    "Traceback on failure"

    corrected_params: Optional[CorrectedParams] = None
    "Model popts and corrected TES parameters"


def find_logalpha_popts(
    chdata: RpFitChanData, cfg: AnalysisCfg
) -> Tuple[np.ndarray, float]:
    """
    Fit Log(alpha') to IV data.

    Returns
    ---------
    popts: np.ndarray
        Optimal parameters for log(alpha') fit.
    pbias_model_offset: float
        Offset between pbias_model from alpha' integration and data [aW]
    """
    smooth_dist = 5
    w_len = 2 * smooth_dist + 1
    w = (1.0 / float(w_len)) * np.ones(w_len)  # window

    rfrac = chdata.iv_R / chdata.iv_Rn
    mask = np.ones_like(rfrac, dtype=bool)
    mask[: chdata.iv_idx_sc + w_len + 2] = False  # Try to cut out SC branch + smoothing
    mask &= (rfrac > cfg.rpfit_min_rfrac) * (rfrac < cfg.rpfit_max_rfrac)
    rfrac = np.convolve(rfrac, w, mode="same")
    pbias = np.convolve(chdata.iv_p_tes, w, mode="same")

    rfrac = rfrac[mask]
    pbias = pbias[mask]
    drfrac = np.diff(rfrac, prepend=np.nan)
    dpbias = np.diff(pbias, prepend=np.nan)
    logalpha = np.log(1 / rfrac * drfrac / dpbias)
    logalpha_mask = np.isfinite(logalpha)

    def logalpha_fit_func(logalpha_pars):
        model = model_logalpha(rfrac, *logalpha_pars)
        chi2 = np.sum((logalpha[logalpha_mask] - model[logalpha_mask]) ** 2)
        return chi2

    fitres = minimize(logalpha_fit_func, [4, 10, 1, 1])
    logalpha_popts = fitres.x

    logalpha_model = model_logalpha(rfrac, *fitres.x)

    pbias_model = np.nancumsum(drfrac / (rfrac * np.exp(logalpha_model)))

    def pbias_offset_fit_func(pbias_offset):
        return np.sum((pbias_model + pbias_offset - pbias) ** 2)

    pbias_model_offset = minimize(pbias_offset_fit_func, pbias[0]).x[0]

    return logalpha_popts, pbias_model_offset


def run_correction(chdata: RpFitChanData, cfg: AnalysisCfg) -> CorrectionResults:
    """
    Runs TES param correction procedure.

    Args
    ----
    chdata: RpFitChanData
        IV and Bias Step calibration data
    cfg: AnalysisCfg
        Correction analysis configuration object.
    """
    res = CorrectionResults(chdata)
    try:
        if np.isnan(chdata.iv_Rn):
            raise ValueError("IV Rn is NaN")

        # Find logalpha popts
        logalpha_popts, pbias_model_offset = find_logalpha_popts(chdata, cfg)

        if np.abs(chdata.bs_Rtes / chdata.iv_Rn) < cfg.sc_rfrac_thresh:
            # Cannot perform correction, since detectors are SC
            res.corrected_params = CorrectedParams(
                logalpha_popts=logalpha_popts,
                pbias_model_offset=pbias_model_offset,
                delta_Popt=np.nan,
                corrected_I0=0.0,
                corrected_R0=0.0,
                corrected_Pj=0.0,
                corrected_Si=np.nan,
                loopgain=np.nan,
            )
            res.success = True
            return res

        # compute dPopt by minimizing (dIrat_IV - dIrat_BS) at chosen bias voltage with respect to delta_popt
        dIrat_meas = chdata.bs_meas_dIrat
        Rsh = chdata.Rsh
        Ibias_setpoint = chdata.bs_Ibias
        rfrac = np.linspace(
            cfg.rpfit_min_rfrac, cfg.rpfit_max_rfrac, cfg.rpfit_intg_nsamps
        )
        dr = rfrac[1] - rfrac[0]
        logalpha_model = model_logalpha(rfrac, *logalpha_popts)
        alpha = np.exp(logalpha_model)
        R = rfrac * chdata.iv_Rn
        pbias = np.cumsum(dr / (rfrac * np.exp(logalpha_model))) + pbias_model_offset

        def dIrat_IV(dPopt):
            P_ = pbias - dPopt
            L0_ = alpha * P_
            i_tes_ = np.sqrt(P_ / R)
            i_bias_ = i_tes_ * (R + Rsh) / Rsh
            irat = (1.0 - L0_) / ((1.0 - L0_) + (1.0 + L0_) * R / Rsh)
            return np.interp(Ibias_setpoint, i_bias_, irat)

        def fit_func(dPopt):
            diff = (dIrat_IV(dPopt) - dIrat_meas) ** 2
            return diff if not np.isnan(diff) else np.inf

        dPopt_fitres = minimize(fit_func, [0])
        if not dPopt_fitres.success:
            raise RuntimeError("dPopt fit failed")

        dPopt = dPopt_fitres.x[0]

        # Adjust IV parameter curves with delta_Popt, and find params at Ibias setpoint
        pbias -= dPopt
        ok = (pbias > 0.0) * (R > Rsh)
        pbias = pbias[ok]
        R = R[ok]
        L0 = alpha[ok] * pbias
        L = (R - Rsh) / (R + Rsh) * L0
        Ites = np.sqrt(pbias / R)
        Ibias = Ites * (R + Rsh) / Rsh
        Si = -1 / (Ites * (R - Rsh)) * (L / (L + 1))

        res.corrected_params = CorrectedParams(
            logalpha_popts=logalpha_popts,
            pbias_model_offset=pbias_model_offset,
            delta_Popt=dPopt,
            corrected_I0=np.interp(Ibias_setpoint, Ibias, Ites).item(),
            corrected_R0=np.interp(Ibias_setpoint, Ibias, R).item(),
            corrected_Pj=np.interp(Ibias_setpoint, Ibias, pbias).item(),
            corrected_Si=np.interp(Ibias_setpoint, Ibias, Si).item(),
            loopgain=np.interp(Ibias_setpoint, Ibias, L0).item(),
        )
        res.success = True
    except Exception:
        if cfg.raise_exceptions:
            raise
        else:
            res.traceback = traceback.format_exc()
            res.success = False

    return res


def run_corrections_parallel(
        iva: IVAnalysis, bsa: BiasStepAnalysis, cfg: AnalysisCfg, nprocs=None,
        executor=None, as_completed_callable=None) -> List[CorrectionResults]:
    """
    Runs correction procedure in parallel for all channels in IV and BSA object
    """

    nchans = iva.nchans
    pb = trange(nchans, disable=(not cfg.show_pb))
    results = []

    if nprocs is None:
        nprocs = cfg.default_nprocs

    # Error handler
    def errback(e):
        raise e

    # Create executor (optionally externally provided)
    if executor is None:
        executor = ProcessPoolExecutor(max_workers=nprocs)
        as_completed_callable = as_completed
        close_executor = True
    else:
        close_executor = False

    try:
        futures = []
        for idx in range(nchans):
            chdata = RpFitChanData.from_data(
                iva, bsa, iva.bands[idx], iva.channels[idx]
            )
            future = executor.submit(run_correction, chdata, cfg)
            futures.append(future)

        for future in as_completed_callable(futures):
            try:
                res = future.result()
                results.append(res)
            except Exception as e:
                errback(e)

            futures.remove(future)
            pb.update(1)

        pb.close()

    finally:
        if close_executor:
            executor.shutdown(wait=True, cancel_futures=True)

    return results

def compute_psats(iva: IVAnalysis, cfg: AnalysisCfg) -> Tuple[np.ndarray, np.ndarray]:
    """
    Re-computes Psat for an IVAnalysis object. Will save results to iva.p_sat.
    This assumes i_tes, v_tes, and r_tes have already been calculated. This will
    not modify the original IVAnalysis object.

    Args
    ----
    iva2 : Dictionary
        Dictionary built from original IV Analysis .npy file
    psat_level : float
        R_frac level for which Psat is defined. If 0.9, Psat will be the
        power on the TES when R_frac = 0.9.

    Returns
    -------
    p_sat : np.ndarray
        Array of length (nchan) with the p-sat computed for each channel (W)
    psat_cross_idx : np.ndarray
        Array of indices at which the psat level is crossed for each channel
    """
    # calculates P_sat as P_TES when Rfrac = psat_level
    # if the TES is at 90% R_n more than once, just take the first crossing
    psats = np.full(iva.nchans, np.nan)
    psat_cross_idx = np.full(iva.nchans, -1)

    for i in range(iva.nchans):
        R = iva.R[i]
        R_n = iva.R_n[i]
        p_tes = iva.p_tes[i]

        if np.isnan(R_n):
            continue

        cross_idx = np.where(R / R_n > cfg.psat_level)[0]
        if len(cross_idx) == 0:
            continue

        # Takes cross-index to be the first time Rfrac crosses psat_level
        cross_idx = cross_idx[0]
        if cross_idx == 0:
            continue

        psat_cross_idx[i] = cross_idx
        try:
            psat = interp1d(
                R[cross_idx - 1 : cross_idx + 1] / R_n,
                p_tes[cross_idx - 1 : cross_idx + 1],
            )(cfg.psat_level)
        except ValueError:
            continue
        psats[i] = psat

    return psats, psat_cross_idx


def compute_si(iva: IVAnalysis) -> np.ndarray:
    """
    Recalculates responsivity using the thevenin equivalent voltage. This will
    not modify the original IVAnalysis object.

    Args
    ----
    iva : IVAnalysis
        IVAnalysis object for which you want to compute Si. This should already
        have items like R, R_n, R_L, i_tes, v_tes, computed.

    Returns
    -------
    si : np.ndarray
        Array of length (nchan, nbiases) with  the responsivity as a fn of
        thevenin equivalent voltage for each channel (V^-1).
    """
    si_all = np.full(iva.si.shape, np.nan)

    smooth_dist = 5
    w_len = 2 * smooth_dist + 1
    w = (1.0 / float(w_len)) * np.ones(w_len)  # window

    v_thev_smooth = np.convolve(iva.v_thevenin, w, mode="same")
    dv_thev = np.diff(v_thev_smooth)

    for i in range(iva.nchans):
        sc_idx = iva.idxs[i, 0]

        if np.isnan(iva.R_n[i]) or sc_idx == -1:
            continue

        # Running average
        i_tes_smooth = np.convolve(iva.i_tes[i], w, mode="same")
        v_tes_smooth = np.convolve(iva.v_tes[i], w, mode="same")
        r_tes_smooth = v_tes_smooth / i_tes_smooth

        R_L = iva.R_L[i]

        # Take derivatives
        di_tes = np.diff(i_tes_smooth)
        dv_tes = np.diff(v_tes_smooth)
        R_L_smooth = np.ones(len(r_tes_smooth - 1)) * R_L
        R_L_smooth[:sc_idx] = dv_tes[:sc_idx] / di_tes[:sc_idx]
        r_tes_smooth_noStray = r_tes_smooth - R_L_smooth
        i0 = i_tes_smooth[:-1]
        r0 = r_tes_smooth_noStray[:-1]
        rL = R_L_smooth[:-1]
        beta = 0.0

        # artificially setting rL to 0 for now,
        # to avoid issues in the SC branch
        # don't expect a large change, given the
        # relative size of rL to the other terms
        # rL = 0

        # Responsivity estimate, derivation done here by MSF
        # https://www.overleaf.com/project/613978cb38d9d22e8550d45c
        si = -(1.0 / (i0 * r0 * (2 + beta))) * (
            1 - ((r0 * (1 + beta) + rL) / (dv_thev / di_tes))
        )
        si[:sc_idx] = np.nan
        si_all[i, :-1] = si

    return si_all


def recompute_ivpars(iva: IVAnalysis, cfg: AnalysisCfg) -> IVAnalysis:
    """
    Takes in an IV Analysis object and analysis cfg params, and recomputes
    TES voltage, current, bias power, resistance, responsivity, and saturation
    powers using the corrected TES formulas for R_L and V_thevenin.
    """
    iva_new = IVAnalysis.from_dict(deepcopy(iva.to_dict()))
    R_sh = iva.meta["R_sh"]
    R_bl = iva.meta["bias_line_resistance"]
    iva_new.i_bias = iva_new.v_bias / R_bl
    iva_new.v_thevenin = iva_new.i_bias * R_sh
    iva_new.v_tes = np.full(iva.v_tes.shape, np.nan)
    iva_new.i_tes = np.full(iva.i_tes.shape, np.nan)
    iva_new.p_tes = np.full(iva.p_tes.shape, np.nan)
    iva_new.R = np.full(iva.R.shape, np.nan)
    iva_new.R_n = np.full(iva.R_n.shape, np.nan)
    iva_new.R_L = np.full(iva.R_L.shape, np.nan)

    iva_new.resp
    for i in range(iva.nchans):
        sc_idx = iva.idxs[i, 0]
        nb_idx = iva.idxs[i, 1]

        R: NDArray[np.float] = R_sh * (iva.i_bias / iva.resp[i] - 1)
        R_par: float = np.nanmean(R[1:sc_idx])
        R_n: float = np.nanmean(R[nb_idx:]) - R_par
        R_L: float = R_sh + R_par
        R_tes: NDArray[np.float] = R - R_par

        iva_new.v_tes[i] = iva_new.v_thevenin * (R_tes / (R_tes + R_L))
        iva_new.i_tes[i] = iva_new.v_tes[i] / R_tes
        iva_new.p_tes[i] = iva_new.v_tes[i] ** 2 / R_tes
        iva_new.R[i] = R_tes
        iva_new.R_n[i] = R_n
        iva_new.R_L[i] = R_L

    iva_new.p_sat, iva_new.idxs[:, 2] = compute_psats(iva_new, cfg)
    iva_new.si = compute_si(iva_new)

    return iva_new



import sodetlib as sdl
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle


def compute_tracking_quality(S, f, df, sync):
    """
    Computes the tracking quality parameter from tracking_setup results.
    Tracking quality is a measure of how periodic the tracking response is with
    respect to the flux-ramp. It is defined as the r-squared value between the
    measured frequency response (f+df), and the (f+df) response averaged across
    all flux-ramp periods. If the signal is regular with respect to flux-ramp,
    the true freq response will be similar to the average freq response,
    resulting in a high r-squared value. If the signal is irregular, the true
    and average freq response will differ, resulting in a lower r-squared
    value. This can be used in conjunction with the f and df peak-to-peak
    values to determine which channels are tracking correctly.

    Args
    ------
    S : SmurfControl
        Pysmurf instance
    f : np.ndarray
        Array of the tracked frequency for each channelk (kHz), as returned by
        tracking_setup
    df : np.ndarray
        Array of the tracked frequency error for each channel (kHz), as
        returned by tracking_setup
    sync : np.ndarray
        Array containing tracking sync flags, as returned by tracking_setup
    """
    sync_idxs = S.make_sync_flag(sync)
    seg_size = np.min(np.diff(sync_idxs))
    nstacks = len(sync_idxs) - 1
    nchans = len(f[0])
    sig = f + df

    fstack = np.zeros((seg_size, nchans))
    for i in range(nstacks):
        si = sync_idxs[i]
        fstack[:seg_size] += (sig)[si:si+seg_size] / nstacks

    # calculates quality of estimate wrt real data
    y_real = (sig)[sync_idxs[0]:sync_idxs[0] + nstacks * seg_size, :]
    y_est = np.vstack([fstack for _ in range(nstacks)])

    # Force these to be the same len in case all segments are not the same size
    y_est = y_est[:len(y_real)]

    with np.errstate(invalid='ignore'):
        sstot = np.sum((y_real - np.nanmean(y_real, axis=0))**2, axis=0)
        ssres = np.sum((y_real - y_est)**2, axis=0)
        r2 = 1 - ssres/sstot

    return r2


class TrackingResults:
    """
    Class for storing, saving, and interpreting results from tracking_setup.
    This class can store tracking results for multiple bands at a time.
    When created, all results array are initialized to be empty. To add
    results from individual bands one at a time, use the ``add_band_data``
    function.

    Args
    ----
    S : SmurfControl
        Pysmurf Instance
    cfg : DetConfig
        DetConfig

    Attributes
    --------------
    meta : dict
        Metadata dictionary
    f_ptp_range : np.ndarray
        Array of len(2) containing the min and max allowed f_ptp. This will
        be pulled from the dev cfg.
    df_ptp_range : np.ndarray
        Array of len(2) containing the min and max allowed df_ptp. This will
        be pulled from the dev cfg.
    r2_min : float
        Min value of r2 for a channel to be considered good
    bands : np.ndarray
        Array of len (nchans) containing the band of each channel
    channels : np.ndarray
        Array of shape (nchans) containing the smurf-channel of each channel
    f : np.ndarray
        Array of shape (nchans, nsamps) containing the tracked freq response
        (kHz) throughout the tracking setup call
    df : np.ndarray
        Array of shape (nchans, nsamps) containing the untracked freq response
        (kHz) throughout the tracking setup call
    sync_idx : np.ndarray
        Array of shape (nchans, num_fr_periods) containing the indices where
        the flux ramp resetted
    r2 : np.ndarray
        Array of shape (nchans) containing the r-squared value computed by
        tracking-quality for each channel
    f_ptp : np.ndarray
        Array of shape (nchans) containing the f_ptp of each channel
    df_ptp : np.ndarray
        Array of shape (nchans) containing the df_ptp of each channel
    is_good : np.ndarray
        Array of shape (nchans) containing True if the channel passes cuts
        and False otherwise
    """
    def __init__(self, *args, **kwargs):
        if len(args) > 0:
            self.initialize(*args, **kwargs)

    def initialize(self, S, cfg):
        self._S = S
        self._cfg = cfg
        self.meta = sdl.get_metadata(S, cfg)
        self.f_ptp_range = np.array(cfg.dev.exp['f_ptp_range'])
        self.df_ptp_range = np.array(cfg.dev.exp['df_ptp_range'])
        self.r2_min = cfg.dev.exp['r2_min']

        self.bands = np.array([], dtype=int)
        self.channels = np.array([], dtype=int)
        self.subband_centers = np.array([], dtype=float)
        self.nchans = 0
        self.ngood = 0

        self.f = None
        self.df = None
        self.tracking_kwargs = [None for _ in range(8)]
        self.sync_idxs = [None for _ in range(8)]
        self.r2 = np.array([], dtype=float)
        self.f_ptp = np.array([], dtype=float)
        self.df_ptp = np.array([], dtype=float)
        self.is_good = np.array([], dtype=bool)

    def add_band_data(self, band, f, df, sync, tracking_kwargs=None):
        """
        Computes tracking-related data based on the tracking response,
        and updates the arrays described in the attributes with channels
        from a new band

        Args
        -----
        band : int
            Band of the data you're adding
        f : np.ndarray
            Tracked freq response as returned by tracking_setup
        df : np.ndarray
            Untracked freq response as returned by tracking_setup
        sync : np.ndarray
            sync arrayas returned by tracking_setup
        """
        if band in self.bands:
            raise ValueError(f"Data for band {band} has already been added!")

        dfptp = np.ptp(df, axis=0)
        m = dfptp != 0

        channels = np.where(m)[0]
        nchans = len(channels)
        bands = np.array([band for _ in channels])
        sb_centers = np.array(
            self._S.get_subband_centers(band, as_offset=False)[1]
        )
        self.channels = np.concatenate((self.channels, channels))
        self.bands = np.concatenate((self.bands, bands))
        self.subband_centers = np.concatenate((
                self.subband_centers, sb_centers[channels]))

        self.tracking_kwargs[band] = tracking_kwargs
        r2 = compute_tracking_quality(self._S, f, df, sync)

        if self.f is not None:
            # It's possible for different tracking setup calls to have a
            # slightly different number of samples, so we have to make sure
            # we cut/elongate results so they can fit into the f/df arrays
            nsamps = len(self.f[0])
            _f = np.full((nchans, nsamps), np.nan)
            _df = np.full((nchans, nsamps), np.nan)
            fi = min(nsamps, len(f))
            _f[:, :fi] = f.T[m, :fi] * 1000
            _df[:, :fi] = df.T[m, :fi] * 1000
            if fi < nsamps:
                # Fill with the last data point to not mess up ptp calcs....
                _f[:, fi:] = _f[:, fi-1][:, None]
                _df[:, fi:] = _df[:, fi-1][:, None]
        else:
            _f = f.T[m] * 1000
            _df = df.T[m] * 1000

        self.r2 = np.concatenate((self.r2, r2[m]))

        f_ptp = np.ptp(_f, axis=1)
        self.f_ptp = np.concatenate((self.f_ptp, f_ptp))

        df_ptp = np.ptp(_df, axis=1)
        self.df_ptp = np.concatenate((self.df_ptp, df_ptp))

        is_good = np.ones_like(r2, dtype=bool)
        self.is_good = np.concatenate((self.is_good, is_good))

        if self.f is None:
            self.f = _f
            self.df = _df
        else:
            nsamps = len(self.f[0])
            self.f = np.concatenate((self.f, _f[:, :nsamps]), axis=0)
            self.df = np.concatenate((self.df, _df[:, :nsamps]), axis=0)

        self.sync_idxs[band] = self._S.make_sync_flag(sync)
        self.nchans += len(_f)
        self.find_bad_chans()

    def find_bad_chans(self, f_ptp_range=None, df_ptp_range=None, r2_min=None):
        """
        Recomputes the ``is_good`` array based on cuts ranges.
        """
        if f_ptp_range is None:
            f_ptp_range = self.f_ptp_range
        if df_ptp_range is None:
            df_ptp_range = self.df_ptp_range
        if r2_min is None:
            r2_min = self.r2_min

        f0, f1 = f_ptp_range
        df0, df1 = df_ptp_range

        self.is_good = np.ones_like(self.r2, dtype=bool)
        self.is_good[self.r2 < r2_min] = 0
        self.is_good[~((f0 < self.f_ptp) & (self.f_ptp < f1))] = 0
        self.is_good[~((df0 < self.df_ptp) & (self.df_ptp < df1))] = 0
        self.ngood = np.sum(self.is_good)

    def save(self, path=None):
        saved_fields = [
            'meta', 'bands', 'channels',
            'f_ptp_range', 'df_ptp_range', 'r2_min',
            'f', 'df', 'sync_idxs', 'r2', 'is_good',
            'tracking_kwargs', 'f_ptp', 'df_ptp',
        ]
        data = {}
        for field in saved_fields:
            data[field] = getattr(self, field)
        if path is None:
            path = sdl.make_filename(self._S, 'tracking_results.npy')

        self.filepath = path
        np.save(path, data, allow_pickle=True)
        if self._S is not None:
            self._S.pub.register_file(path, 'tracking_data', format='npy')

    @classmethod
    def load(cls, path):
        self = cls()
        for k, v in np.load(path, allow_pickle=True).item().items():
            setattr(self, k, v)
        return self


def plot_tracking_channel(tr, idx, show_text=True):
    """
    Plots single tracking channel from results
    """
    f = tr.f[idx]
    df = tr.df[idx]
    fig, ax = plt.subplots(figsize=(12, 4))
    band = int(tr.bands[idx])
    channel = int(tr.channels[idx])

    ax.set_title(f"Band {band} Channel {channel}")
    ax.plot(df+f, color='grey', alpha=0.8, label='Freq Response')
    ax.plot(f, label='Tracked Freq')
    ax.set_xticks([])
    if hasattr(tr, 'subband_centers'):
        ax.set_ylabel(f"Freq offset from {tr.subband_centers[idx]:0.2f} [kHz]")
    else:
        ax.set_ylabel(f"Freq offset [kHz]")
    ax.legend(loc='upper left')

    if show_text:
        txt = '\n'.join([
            f'fptp = {tr.f_ptp[idx]:0.2f}',
            f'dfptp = {tr.df_ptp[idx]:0.2f}',
            f'r2 = {tr.r2[idx]:0.2f}',
            f'is_good = {tr.is_good[idx]}',
        ])
        bbox = dict(facecolor='white', alpha=0.8)
        ax.text(0.02, 0.1, txt, transform=ax.transAxes, bbox=bbox)
    for s in tr.sync_idxs[band]:
        ax.axvline(s, color='grey', ls='--')
    return fig, ax


def plot_tracking_summary(tr):
    """
    Plots summary of tracking results
    """
    # fig, axes = plt.subplots(figsize=(12, 8))
    mosaic = """
    BBBB.
    AAAAC
    AAAAC
    AAAAC
    AAAAC
    """
    gs = dict(hspace=0, wspace=0)
    fig, axes = plt.subplot_mosaic(mosaic, figsize=(12, 8), gridspec_kw=gs)
    f0, f1 = tr.f_ptp_range
    df0, df1 = tr.df_ptp_range

    m = tr.is_good
    ax = axes['A']
    ax.scatter(tr.f_ptp[m], tr.df_ptp[m], marker='.', alpha=0.8)
    ax.scatter(tr.f_ptp[~m], tr.df_ptp[~m], marker='.', color='red', alpha=0.2)
    xlim = (-5, f1 * 1.3)
    ylim = (-5, df1 * 1.3)
    ax.set_xlim(*xlim)
    ax.set_ylim(*ylim)
    rect = Rectangle((f0, df0), f1-f0, df1-df0, fc='green', alpha=0.2,
                     linestyle='--', ec='black', lw=3)
    ax.add_patch(rect)
    ax.set_ylabel(r"$df_\mathrm{ptp}$ (kHz)", fontsize=20)
    ax.set_xlabel(r"$f_\mathrm{ptp}$ (kHz)", fontsize=20)

    text = f"{tr.ngood} / {tr.nchans} good tracking channels"
    bbox = dict(fc='wheat', alpha=0.8)
    ax.text(0.02, 0.9, text, transform=ax.transAxes, bbox=bbox, fontsize=16)

    ax = axes['B']
    ax.set_xticks([])
    ax.hist(tr.f_ptp, range=xlim, bins=40)
    ax.set_xlim(*xlim)
    ax.axvspan(f0, f1, color='green', alpha=0.2)

    ax = axes['C']
    ax.set_yticks([])
    ax.hist(tr.df_ptp, range=ylim, bins=40, orientation='horizontal')
    ax.set_ylim(*ylim)
    ax.axhspan(df0, df1, color='green', alpha=0.2)

    return fig, ax


def disable_bad_chans(S, tr, bands=None, **kwargs):
    """
    Disables cut channels based on a TrackingResults object.
    """
    tr.find_bad_chans(**kwargs)
    if bands is None:
        bands = np.arange(8)

    bands = np.atleast_1d()
    for b in bands:
        m = tr.bands == b
        asa = S.get_amplitude_scale_array(b)
        fea = S.get_feedback_enable_array(b)
        for good, c in zip(tr.is_good[m], tr.channels[m]):
            if not good:
                asa[c] = 0
                fea[c] = 0
        S.set_amplitude_scale_array(b, asa)
        S.set_feedback_enable_array(b, asa)


@sdl.set_action()
def setup_tracking_params(S, cfg, bands, update_cfg=True, show_plots=False):
    """
    Sets up tracking parameters by determining correct frac-pp and lms-freq
    for each band.

    Args
    -----
    S : SmurfControl
        Pysmurf instance
    cfg : DetConfig
        DetConfig instance
    bands : np.ndarray, int
        Band or list of bands to run on
    update_cfg : bool
        If true, will update the device cfg and save the file.
    show_plots : bool
        If true, will show summary plots
    """

    bands = np.atleast_1d(bands)
    exp = cfg.dev.exp

    res = TrackingResults(S, cfg)

    tk = {
        'reset_rate_khz': exp['flux_ramp_rate_khz'],
        'make_plot': False, 'show_plot': False, 'channel': [],
        'nsamp': 2**18, 'return_data': True,
        'feedback_start_frac': exp['feedback_start_frac'],
        'feedback_end_frac': exp['feedback_end_frac'],
    }
    for band in bands:
        sdl.pub_ocs_log(S, f"Setting up trackng params: band {band}")
        bcfg = cfg.dev.bands[band]
        tk.update({
            'lms_freq_hz':         None,
            'meas_lms_freq':       True,
            'fraction_full_scale': exp['init_frac_pp'],
            'reset_rate_khz':      exp['flux_ramp_rate_khz'],
            'lms_gain': bcfg['lms_gain'],
            'feedback_gain': bcfg['feedback_gain'],
        })

        f, df, sync = S.tracking_setup(band, **tk)
        tr = TrackingResults(S, cfg)
        tr.add_band_data(band, f, df, sync)
        asa_init = S.get_amplitude_scale_array(band)
        disable_bad_chans(S, tr, bands=band, r2_min=0.95)

        # Calculate trracking parameters
        S.tracking_setup(band, **tk)
        lms_meas = S.lms_freq_hz[band]
        lms_freq = exp['nphi0'] * tk['reset_rate_khz'] * 1e3
        frac_pp = tk['fraction_full_scale'] * lms_freq / lms_meas

        # Re-enables all channels and re-run tracking setup with correct params
        S.set_amplitude_scale_array(band, asa_init)
        tk.update({
            'meas_lms_freq': False,
            'fraction_full_scale': frac_pp,
            'lms_freq_hz': lms_freq,
        })

        f, df, sync = S.tracking_setup(band, **tk)
        res.add_band_data(band, f, df, sync, tracking_kwargs=tk)

        # Print tracking params
        d = {
            'frac_pp': frac_pp,
            'lms_freq_hz': lms_freq,
        }
        S.log(f"Tracking params found for band {band}: {d}")

        # Update det config
        if update_cfg:
            cfg.dev.update_band(band, d, update_file=True)

    res.save()

    S.log("Running relock tracking setup...")
    res = relock_tracking_setup(S, cfg, bands=bands, show_plots=show_plots)

    return res


def _get_fixed_tone_channels(S, cfg, bands=None):
    """
    Identify channels assigned to fixed tones by checking state of amplitude and
    feedback arrays.

    Args
    ----
    S : SmurfControl
        Pysmurf instance
    cfg : DetConfig
        Det config instance
    bands : list of int
        Bands to operate on. Get from config by default.
    """

    if bands is None:
        bands = cfg.dev.exp['active_bands']

    # infer from amplitude and feedback state
    ft_ind = {}
    for band in bands:
        amp = S.get_amplitude_scale_array(band)
        feedback = S.get_feedback_enable_array(band)
        ft_ind[band] = np.where((amp != 0) & (feedback == 0))[0]

    return ft_ind


def _restore_fixed_tone_channels(S, chan_per_band: dict):
    """
    Ensure feedback is disabled on channels that are assigned fixed tones.

    Args
    ----
    S : SmurfControl
        Pysmurf instance
    chan_per_band : dict
        List of channels to disable feedback on, given as a dictionary indexed by band number.
    """

    # infer from amplitude and feedback state
    for band, chan in chan_per_band.items():
        feedback = S.get_feedback_enable_array(band)
        feedback[chan] = 0
        S.set_feedback_enable_array(band, feedback)


@sdl.set_action()
def relock_tracking_setup(S, cfg, bands=None, reset_rate_khz=None, nphi0=None,
                          feedback_gain=None, lms_gain=None, show_plots=False,
                          frac_pp=None): 
    """
    Sets up tracking for smurf. This assumes you already have optimized
    lms_freq and frac-pp for each bands in the device config. This function
    will chose the flux-ramp fraction-full-scale by averaging the optimized
    fractions across the bands you're running on.

    This function also allows you to set reset_rate_khz and nphi0. The
    fraction-full-scale, and lms frequencies of each band will be automatically
    adjusted based on their pre-existing optimized values.

    Additional keyword args specified will be passed to S.tracking_setup.

    Args
    -----
    S : SmurfControl
        Pysmurf instance
    cfg : DetConfig
        Det config instance
    reset_rate_khz : float, optional
        Flux Ramp Reset Rate to set (kHz), defaults to the value in the dev cfg
    nphi0 : int, optional
        Number of phi0's to ramp through. Defaults to the value that was used
        during setup.
    disable_bad_chans : bool
        If true, will disable tones for bad-tracking channels

    Returns
    --------
    res : dict
        Dictionary of results of all tracking-setup calls, with the bands number
        as key.
    """
    if bands is None:
        bands = cfg.dev.exp['active_bands']
    bands = np.atleast_1d(bands)

    nbands = len(bands)
    exp = cfg.dev.exp

    # check if any channels are assigned fixed tones
    fixed_tones = _get_fixed_tone_channels(S, cfg, bands)

    # Arrays containing the optimized tracking parameters for each band
    frac_pp0 = np.zeros(nbands)
    lms_freq0 = np.zeros(nbands)  # Hz
    reset_rate_khz0 = exp['flux_ramp_rate_khz']
    init_nphi0 = exp['nphi0']

    for i, b in enumerate(bands):
        bcfg = cfg.dev.bands[b]
        frac_pp0[i] = bcfg['frac_pp']
        lms_freq0[i] = bcfg['lms_freq_hz']

    # Choose frac_pp to be the mean of all running bands.
    # This is the frac-pp at the flux-ramp-rate used for optimization
    fpp0 = np.mean(frac_pp0)
    lms_freq0 *= fpp0 / frac_pp0
    
    S.log(f"Using frac-pp={fpp0}")

    # Adjust fpp, lms_freq, and flux-ramp-rate depending on desired
    # flux-ramp-rate and nphi0
    fpp, lms_freqs = fpp0, lms_freq0
    if nphi0 is not None:
        fpp *= nphi0 / init_nphi0
        lms_freqs *= fpp / fpp0
    if reset_rate_khz is not None:
        lms_freqs *= reset_rate_khz / reset_rate_khz0
    else:
        reset_rate_khz = reset_rate_khz0

    if frac_pp is not None:
        lms_freqs *=  frac_pp / fpp
        fpp = frac_pp

    res = TrackingResults(S, cfg)
    tk = {
        'reset_rate_khz': reset_rate_khz, 'fraction_full_scale': fpp,
        'make_plot': False, 'show_plot': False, 'channel': [],
        'nsamp': 2**18, 'return_data': True,
        'feedback_start_frac': exp['feedback_start_frac'],
        'feedback_end_frac': exp['feedback_end_frac'],
    }


    for i, band in enumerate(bands):
        bcfg = cfg.dev.bands[band]
        tk.update({
            'lms_freq_hz': lms_freqs[i],
            'lms_gain': bcfg['lms_gain'],
            'feedback_gain': bcfg['feedback_gain'],
        })

        if lms_gain is not None:
            tk['lms_gain'] = lms_gain
        if feedback_gain is not None:
            tk['feedback_gain'] = feedback_gain

        f, df, sync = S.tracking_setup(band, **tk)
        res.add_band_data(band, f, df, sync, tracking_kwargs=tk)

    res.find_bad_chans()
    res.save()

    # ensure channels with assigned fixed tones have feedback disabled
    _restore_fixed_tone_channels(S, fixed_tones)

    is_interactive = plt.isinteractive()
    try:
        if not show_plots:
            plt.ioff()
        fig, ax = plot_tracking_summary(res)
        path = sdl.make_filename(S, 'tracking_results.png', plot=True)
        fig.savefig(path)
        S.pub.register_file(path, 'tracking_summary', plot=True, format='png')
        if not show_plots:
            plt.close(fig)
    finally:
        if is_interactive:
            plt.ion()

    return res



import time
import os
import traceback
import numpy as np
import sodetlib as sdl
import matplotlib.pyplot as plt
import scipy.optimize
from sodetlib.operations import iv
from typing import Dict, Any, Optional
from dataclasses import dataclass, asdict
from copy import deepcopy

np.seterr(all='ignore')


def _play_tes_bipolar_waveform(S, bias_group, waveform, do_enable=True,
                                 continuous=True, **kwargs):
    """
    Play a bipolar waveform on the bias group.
    Args
    ----
    bias_group : int
                The bias group
    waveform : float array
                The waveform the play on the bias group.
    do_enable : bool, optional, default True
                Whether to enable the DACs (similar to what is required
                for TES bias).
    continuous : bool, optional, default True
                Whether to play the TES waveform continuously.
    """
    bias_order = S.bias_group_to_pair[:,0]

    dac_positives = S.bias_group_to_pair[:,1]
    dac_negatives = S.bias_group_to_pair[:,2]

    dac_idx = np.ravel(np.where(bias_order == bias_group))

    dac_positive = dac_positives[dac_idx][0]
    dac_negative = dac_negatives[dac_idx][0]

    # https://confluence.slac.stanford.edu/display/SMuRF/SMuRF+firmware#SMuRFfirmware-RTMDACarbitrarywaveforms
    # Target the two bipolar DACs assigned to this bias group:
    S.set_dac_axil_addr(0, dac_positive)
    S.set_dac_axil_addr(1, dac_negative)

    # Must enable the DACs (if not enabled already)
    if do_enable:
        S.set_rtm_slow_dac_enable(dac_positive, 2, **kwargs)
        S.set_rtm_slow_dac_enable(dac_negative, 2, **kwargs)

    # Load waveform into each DAC's LUT table.  Opposite sign so
    # they combine coherenty
    S.set_rtm_arb_waveform_lut_table(0, waveform)
    S.set_rtm_arb_waveform_lut_table(1, -waveform)

    # Enable waveform generation (1=on both DACs)
    S.set_rtm_arb_waveform_enable(3)

    # Continous mode to play the waveform continuously
    if continuous:
        S.set_rtm_arb_waveform_continuous(1)
    else:
        S.set_rtm_arb_waveform_continuous(0)

def _make_step_waveform(S, step_dur, step_voltage, dc_voltage):
    """
    Returns a waveform for a bias step. Waveform will contain step-up
    to (dc_voltage + step_voltage) for `step_dur` seconds, and then a
    step-down to (dc_voltage) for `step_dur` seconds.

    Args
    -----
    S : SmurfControl
        Pysmurf instance
    step_dur : float
        Duration of each step (sec)
    step_voltage : float
        Amplitude of step (V)
    dc_voltage : float
        DC voltage of step

    Returns
    ---------
    sig : np.ndarray
        Waveform to use
    timer_size : float
        Waveform timer size
    """
    # Setup waveform
    sig = np.ones(2048)
    sig *= dc_voltage / (2*S._rtm_slow_dac_bit_to_volt)
    sig[1024:] += step_voltage / (2*S._rtm_slow_dac_bit_to_volt)
    # Use twice the step dur because the signal contains two steps
    timer_size = int(step_dur*2/(6.4e-9 * 2048))
    return sig, timer_size

def play_bias_steps_waveform(S, cfg, bias_group, step_duration, step_voltage,
                             num_steps=5, dc_bias=None):
    """
    Plays bias steps  on a single bias line using the Arb waveform generator

    Args
    ----
    S:
        Pysmurf control instance
    cfg:
        DetConfig instance
    bias_group: int
        Bias groups to play bias step on. Defaults to all 12
    step_duration: float
        Duration of each step in sec
    step_voltage : float
        Step size (volts)
    num_steps: int
        Number of bias steps
    dacs : str
        Which group of DACs to play bias-steps on. Can be 'pos',
        'neg', or 'both'
    """
    if dc_bias is None:
        dc_bias = S.get_tes_bias_bipolar(bias_group)
    sig, timer_size = _make_step_waveform(S, step_duration, step_voltage, dc_bias)
    S.set_rtm_arb_waveform_timer_size(timer_size, wait_done=True)
    _play_tes_bipolar_waveform(S, bias_group, sig)
    start_time = time.time()
    time.sleep(step_duration * (num_steps+1))
    stop_time = time.time()
    S.set_rtm_arb_waveform_enable(0)
    S.set_tes_bias_bipolar(bias_group, dc_bias)
    return start_time, stop_time


def play_bias_steps_dc(S, cfg, bias_groups, step_duration, step_voltage,
                       num_steps=5, dacs='pos'):
    """
    Plays bias steps on a group of bias groups stepping with only one DAC

    Args:
        S:
            Pysmurf control instance
        cfg:
            DetConfig instance
        bias_group: (int, list, optional)
            Bias groups to play bias step on. Defaults to all 12
        step_duration: float
            Duration of each step in sec
        step_voltage : float
            Step size (volts)
        num_steps: int
            Number of bias steps
        dacs : str
            Which group of DACs to play bias-steps on. Can be 'pos',
            'neg', or 'both'
    """
    if bias_groups is None:
        bias_groups = cfg.dev.exp['active_bgs']
    bias_groups = np.atleast_1d(bias_groups)

    if dacs not in ['pos', 'neg', 'both']:
        raise ValueError("Arg dac must be in ['pos', 'neg', 'both']")

    dac_volt_array_low = S.get_rtm_slow_dac_volt_array()
    dac_volt_array_high = dac_volt_array_low.copy()

    bias_order, dac_positives, dac_negatives = S.bias_group_to_pair.T

    for bg in bias_groups:
        bg_idx = np.ravel(np.where(bias_order == bg))
        dac_positive = dac_positives[bg_idx][0] - 1
        dac_negative = dac_negatives[bg_idx][0] - 1
        if dacs == 'pos':
            dac_volt_array_high[dac_positive] += step_voltage
        elif dacs == 'neg':
            dac_volt_array_high[dac_negative] -= step_voltage
        elif dacs == 'both':
            dac_volt_array_high[dac_positive] += step_voltage / 2
            dac_volt_array_high[dac_negative] -= step_voltage / 2

    start = time.time()
    for _ in range(num_steps):
        S.set_rtm_slow_dac_volt_array(dac_volt_array_high)
        time.sleep(step_duration)
        S.set_rtm_slow_dac_volt_array(dac_volt_array_low)
        time.sleep(step_duration)
    stop = time.time()

    return start, stop


def exp_fit(t, A, tau, b):
    """
    Fit function for exponential falloff in bias steps
    """
    return A * np.exp(-t / tau) + b


class BiasStepAnalysis:
    """
    Container to manage analysis of bias steps taken with the take_bias_steps
    function. The main function is ``run_analysis`` and will do a series of
    analysis procedures to create a biasgroup map and calculate DC detector
    parameters and tau_eff:

     - Loads an axis manager with all the data
     - Finds locations of step edges for each bias group
     - Creates a bg map using the isolated bg step responses
     - Gets detector responses of each step and aligns them based on the time
       of the step
     - Computes DC params R0, I0, Pj, Si from step responses
     - Fits exponential to the average step response and estimates tau_eff.

    Most analysis inputs and products will be saved to a npy file so they can
    be loaded and re-analyzed easily on another computer like simons1.

    To load data from an saved step file, you can run::

        bsa = BiasStepAnalysis.load(<path>)

    Attributes:
        tunefile (path): Path of the tunefile loaded by the pysmurf instance
        high_low_current_ratio (float):
            Ratio of high to low current
        R_sh (float):
            Shunt resistance loaded into pysmurf at time of creation
        pA_per_phi0 (float):
            pA_per_phi0, as loaded in pysmurf at time of creation
        rtm_bit_to_volt (float):
            Conversion between bias dac bit and volt
        bias_line_resistance (float):
            Bias line resistance loaded in pysmurf at time of creation
        high_current_mode (bool):
            If high-current-mode was used
        stream_id (string):
            stream_id of the streamer this was run on.
        sid (int):
            Session-id of streaming session
        start, stop (float):
            start and stop time of all steps
        edge_idxs (array(ints) of shape (nbgs, nsteps)):
            Array containing indexes (wrt axis manager) of bias group steps
            for each bg
        edge_signs (array(+/-1) of shape (nbgs, nsteps)):
            Array of signs of each step, denoting whether step is rising or
            falling
        bg_corr (array (float) of shape (nchans, nbgs)):
            Bias group correlation array, stating likelihood that a given
            channel belongs on a given bias group determined from the isolated
            steps
        bgmap (array (int) of shape (nchans)):
            Map from readout channel to assigned bias group. -1 means not
            assigned (that the assignment threshold was not met for any of the
            12 bgs)
        abs_chans (array (int) of shape (nchans)):
            Array of the absolute smurf channel number for each channel in the
            axis manager.
        resp_times (array (float) shape (nbgs, npts)):
            Shared timestamps for each of the step responses in <step_resp> and
            <mean_resp> with respect to the bg-step location, with the step
            occuring at t=0.
        mean_resp (array (float) shape (nchans, npts)):
            Step response averaged accross all bias steps for a given channel
            in Amps.
        step_resp (array (float) shape (nchans, nsteps, npts)):
            Each individual step response for a given channel in amps
        Ibias (array (float) shape (nbgs)):
            DC bias current of each bias group (amps)
        Vbias:
            DC bias voltage of each bias group (volts in low-current mode)
        dIbias (array (float) shape (nbgs)):
            Step current for each bias group (amps)
        dVbias (array (float) shape (nbgs)):
            Step voltage for each bias group (volts in low-current mode)
        dItes (array (float) shape (nchans)):
            Array of tes step heigh for each channel (amps)
        R0 (array (float) shape (nchans)):
            Computed TES resistances for each channel (ohms)
        I0 (array (float) shape (nchans)):
            Computed TES currents for each channel (amps)
        Pj (array (float) shape (nchans)):
            Bias power computed for each channel (W)
        Si (array (float) shape (nchans)):
            Responsivity computed for each channel (1/V)
        step_fit_tmin (array (float) shape (nchans)):
            Time after bias step to start fitting exponential (sec)
        step_fit_popts (array (float) of shape (nchans, 3)):
            Optimal fit parameters (A, tau, b) for the exponential fit of each
            channel
        step_fit_pcovs (array (float) shape (nchans, 3, 3)):
            Fit covariances for each channel
        tau_eff (array (float) shape (nchans)):
            Tau_eff for each channel (sec). Same as step_fit_popts[:, 1].
        R_n_IV (array (float) shape (nchans)):
            Array of normal resistances for each channel pulled from IV
            in the device cfg.
        Rfrac (array (float) shape (nchans)):
            Rfrac of each channel, determined from R0 and the channel's normal
            resistance.
    """

    def __init__(self, S=None, cfg=None, run_kwargs=None):
        self._S = S
        self._cfg = cfg

        self.am = None
        self.edge_idxs = None
        self.transition_range = None

        if S is not None:
            self.meta = sdl.get_metadata(S, cfg)
            self.stream_id = cfg.stream_id

            if run_kwargs is None:
                run_kwargs = {}
            self.run_kwargs = run_kwargs
            self.high_current_mode = run_kwargs.get("high_current_mode", True)


    @classmethod
    def from_dict(cls, data) -> 'BiasStepAnalysis':
        self = cls()
        for k, v in data.items():
            setattr(self, k, v)
        return self

    def to_dict(self) -> Dict[str, Any]:
        data = {}
        saved_fields = [
            # Run data and metadata
            'bands', 'channels', 'sid', 'meta', 'run_kwargs', 'start', 'stop',
            'high_current_mode',
            # Bgmap data
            'bgmap', 'polarity',
            # Step data and fits
            'resp_times', 'mean_resp', 'step_resp',
            # Step fit data
            'step_fit_tmin', 'step_fit_popts', 'step_fit_pcovs',
            'tau_eff',
            # Det param data
            'transition_range', 'Ibias', 'Vbias', 'dIbias', 'dVbias', 'dItes',
            'R0', 'I0', 'Pj', 'Si',
            # From IV's
            'R_n_IV', 'Rfrac',
        ]

        for f in saved_fields:
            if not hasattr(self, f):
                print(f"WARNING: field {f} does not exist... "
                      "defaulting to None")
            data[f] = getattr(self, f, None)
        return data

    def save(self, path=None) -> None:
        data = self.to_dict()
        if path is not None:
            np.save(path, data, allow_pickle=True)
            self.filepath = path
        else:
            self.filepath = sdl.validate_and_save(
                'bias_step_analysis.npy', data, S=self._S, cfg=self._cfg,
                make_path=True
            )

    @classmethod
    def load(cls, filepath) -> 'BiasStepAnalysis':
        self = cls.from_dict(np.load(filepath, allow_pickle=True).item())
        self.filepath = filepath
        return self

    def run_analysis(
            self, create_bg_map=False, assignment_thresh=0.3, save_bg_map=True,
            arc=None, base_dir='/data/so/timestreams', step_window=0.03, fit_tmin=1.5e-3,
            transition=None, R0_thresh=30e-3, save=False, bg_map_file=None, fit_tau=True,
    ):
        """
        Runs the bias step analysis.

        Parameters:
            create_bg_map (bool):
                If True, will create a bg map from the step data. If False,
                will use the bgmap from the device cfg
            assignment_thresh (float):
                Correlation threshold for which channels should be assigned to
                particular bias groups.
            save_bg_map (bool):
                If True, will save the created bgmap to disk and set it as
                the bgmap path in the device cfg.
            arc (optional, G3tSmurf):
                G3tSmurf archive. If specified, will attempt to load
                axis-manager using archive instead of sid.
            base_dir (optiional, str):
                Base directory where timestreams are stored. Defaults to
                /data/so/timestreams.
            step_window (float):
                Time after the bias step (in seconds) to use for the analysis.
            fit_tmin (float):
                DEPRECATED!! do not use.
            transition: (tuple, bool, optional)
                DEPRECATED!! do not use.
            R0_thresh (float):
                Any channel with resistance greater than R0_thresh will be
                unassigned from its bias group under the assumption that it's
                crosstalk
            save (bool):
                If true will save the analysis to a npy file.
            bg_map_file (optional, path):
                If create_bg_map is false and this file is not None, use this file
                to load the bg_map.
            fit_tau (bool):
                If True, perform the time constant fit. If False, skip this.
        """
        self._load_am(arc=arc, base_dir=base_dir)
        self._find_bias_edges()
        if create_bg_map:
            self._create_bg_map(assignment_thresh=assignment_thresh,
                                save_bg_map=save_bg_map)
        elif bg_map_file is not None:
            self.bgmap, self.polarity = sdl.load_bgmap(
                self.bands, self.channels, bg_map_file)
        else:
            self.bgmap, self.polarity = sdl.load_bgmap(
                self.bands, self.channels, self.meta['bgmap_file'])

        self._get_step_response(step_window=step_window)
        self._compute_dc_params(R0_thresh=R0_thresh)

        # Load R_n from IV
        self.R_n_IV = np.full(self.nchans, np.nan)
        # Rfrac determined from R0 and R_n
        self.Rfrac = np.full(self.nchans, np.nan)
        if self.meta['iv_file'] is not None:
            if os.path.exists(self.meta['iv_file']):
                iva = iv.IVAnalysis.load(self.meta['iv_file'])
                chmap = sdl.map_band_chans(
                    self.bands, self.channels, iva.bands, iva.channels
                )
                self.R_n_IV = iva.R_n[chmap]
                self.R_n_IV[chmap == -1] = np.nan
                self.Rfrac = self.R0 / self.R_n_IV

        if create_bg_map and save_bg_map and self._S is not None:
            # Write bgmap after compute_dc_params because bg-assignment
            # will be un-set if resistance estimation is too high.
            ts = str(int(time.time()))
            data = {
                'bands': self.bands,
                'channels': self.channels,
                'sid': self.sid,
                'meta': self.meta,
                'bgmap': self.bgmap,
                'polarity': self.polarity,
            }
            path = os.path.join('/data/smurf_data/bias_group_maps',
                                ts[:5],
                                self.meta['stream_id'],
                                f'{ts}_bg_map.npy')
            os.makedirs(os.path.dirname(path), exist_ok=True)
            sdl.validate_and_save(path, data, S=self._S, cfg=self._cfg,
                                  register=True, make_path=False)
            self._cfg.dev.update_experiment({'bgmap_file': path},
                                            update_file=True)

        if fit_tau:
            self._fit_tau_effs()

        if save:
            self.save()

    def _load_am(self, arc=None, base_dir='/data/so/timestreams', fix_timestamps=True):
        """
        Attempts to load the axis manager from the sid or return one that's
        already loaded. Also sets the `abs_chans` array.

        TODO: adapt this to work on with a general G3tSmurf archive if
        supplied.
        """
        if self.am is None:
            if arc:
                self.am = arc.load_data(self.start, self.stop, stream_id=self.meta['stream_id'])
            else:
                self.am = sdl.load_session(self.meta['stream_id'], self.sid,
                                           base_dir=base_dir)

            # Fix up timestamp jitter from timestamping in software
            if fix_timestamps:
                fsamp, t0 = np.polyfit(self.am.primary['FrameCounter'],
                                       self.am.timestamps, 1)
                self.am.timestamps = t0 + self.am.primary['FrameCounter']*fsamp
            if "det_info" in self.am:
                self.bands = self.am.det_info.smurf.band
                self.channels = self.am.det_info.smurf.channel
            else:
                self.bands = self.am.ch_info.band
                self.channels = self.am.ch_info.channel
            self.abs_chans = self.bands*512 + self.channels
            self.nbgs = len(self.am.biases)
            self.nchans = len(self.am.signal)
        return self.am

    def _find_bias_edges(self, am=None):
        """
        Finds sample indices and signs of bias steps in timestream.

        Returns
        --------
            edge_idxs: list
                List containing the edge sample indices for each bias group.
                There are n_biaslines elements, each one a np.ndarray
                contianing a sample idx for each edge found

            edge_signs: list
                List with the same shape as edge_idxs, containing +/-1
                depending on whether the edge is rising or falling
        """
        if am is None:
            am = self.am

        edge_idxs = [[] for _ in am.biases]
        edge_signs = [[] for _ in am.biases]

        for bg, bias in enumerate(am.biases):
            idxs = np.where(np.diff(bias) != 0)[0]
            signs = np.sign(np.diff(bias)[idxs])

            # Delete double steps
            doubled_idxs = np.where(np.diff(signs) == 0)[0]
            idxs = np.delete(idxs, doubled_idxs)
            signs = np.delete(signs, doubled_idxs)

            edge_idxs[bg] = idxs
            edge_signs[bg] = signs

        self.edge_idxs = edge_idxs
        self.edge_signs = edge_signs

        return edge_idxs, edge_signs

    def _create_bg_map(self, step_window=0.03, assignment_thresh=0.3,
                       save_bg_map=True):
        """
        Creates a bias group mapping from the bg step sweep. The step sweep
        goes down and steps each bias group one-by-one up and down twice. A
        bias group correlation factor is computed by integrating the diff of
        the TES signal times the sign of the step for each bg step. The
        correlation factors are normalized such that the sum across bias groups
        for each channel is 1, and an assignment is made if the normalized
        bias-group correlation is greater than some threshold.

        Saves:
            bg_corr (np.ndarray):
                array of shape (nchans, nbgs) that contain the correlation
                factor for each chan/bg combo (normalized st the sum is 1).
            bgmap (np.ndarray):
                Array of shape (nchans) containing the assigned bg of each
                channel, or -1 if no assigment could be determined
        """
        am = self._load_am()
        if self.edge_idxs is None:
            self._find_bias_edges()

        fsamp = np.nanmean(1./np.diff(am.timestamps))
        npts = int(fsamp * step_window)

        nchans = len(am.signal)
        nbgs = len(am.biases)
        bgs = np.arange(nbgs)
        bg_corr = np.zeros((nchans, nbgs))

        for bg in bgs:
            for i, ei in enumerate(self.edge_idxs[bg]):
                s = slice(ei, ei+npts)
                ts = am.timestamps[s]
                if not (self.start < ts[0] < self.stop):
                    continue
                sig = self.edge_signs[bg][i] * am.signal[:, s]
                bg_corr[:, bg] += np.sum(np.diff(sig), axis=1)
        abs_bg_corr = np.abs(bg_corr)
        normalized_bg_corr = (abs_bg_corr.T / np.sum(abs_bg_corr, axis=1)).T
        normalized_bg_corr[np.isnan(normalized_bg_corr)] = 0.
        bgmap = np.nanargmax(normalized_bg_corr, axis=1)
        m = np.max(normalized_bg_corr, axis=1) < assignment_thresh
        bgmap[m] = -1

        # Calculate the sign of each channel
        self.polarity = np.ones(self.nchans, dtype=int)
        for i in range(self.nchans):
            self.polarity[i] = np.sign(bg_corr[i, bgmap[i]])

        self.bg_corr = normalized_bg_corr
        self.bgmap = bgmap

        return self.bgmap

    def _get_step_response(self, step_window=0.03, pts_before_step=20,
                           restrict_to_bg_sweep=False, am=None):
        """
        Finds each channel's response to the bias step by looking at the signal
        in a small window of <npts> around each edge-index.

        Saves:
            resp_times:
                Array of shape (nbgs, npts) containing the shared timestamps
                for channels on a given bias group
            step_resp:
                Array of shape (nchans, nsteps, npts) containing the response
                of each channel in a window around every step
            mean_resp:
                Array of (nchans, npts) containing the averaged step response
                for each channel
        """
        if am is None:
            am = self.am

        fsamp = np.nanmean(1./np.diff(am.timestamps))
        pts_after_step = int(fsamp * step_window)
        nchans = len(am.signal)
        nbgs = len(am.biases)
        npts = pts_before_step + pts_after_step
        n_edges = np.max([len(ei) for ei in self.edge_idxs])

        sigs = np.full((nchans, n_edges, npts), np.nan)
        ts = np.full((nbgs, npts), np.nan)

        A_per_rad = self.meta['pA_per_phi0'] / (2*np.pi) * 1e-12
        for bg in np.unique(self.bgmap):
            if bg == -1:
                continue
            rcs = np.where(self.bgmap == bg)[0]
            for i, ei in enumerate(self.edge_idxs[bg][:-2]):
                if i < 2:
                    continue
                s = slice(ei - pts_before_step, ei + pts_after_step)
                if np.isnan(ts[bg]).all():
                    ts[bg, :] = am.timestamps[s] - am.timestamps[ei]
                sig = self.edge_signs[bg][i] * am.signal[rcs, s] * A_per_rad
                # Subtracts mean of last 10 pts such that step ends at 0
                sigs[rcs, i, :] = (sig.T - np.nanmean(sig[:, -10:], axis=1)).T

        self.resp_times = ts
        self.step_resp = (sigs.T * self.polarity).T
        self.mean_resp = (np.nanmean(sigs, axis=1).T * self.polarity).T

        return ts, sigs

    def _compute_R0_I0_Pj(self):
        """
        Computes the DC params R0 I0 and Pj
        """
        Ib = self.Ibias[self.bgmap]
        dIb = self.dIbias[self.bgmap]
        dItes = self.dItes
        Ib[self.bgmap == -1] = np.nan
        dIb[self.bgmap == -1] = np.nan
        dIrat = dItes / dIb

        R_sh = self.meta["R_sh"]

        I0 = np.zeros_like(dIrat)
        I0_nontransition = Ib * dIrat
        I0_transition = Ib * dIrat / (2 * dIrat - 1)
        I0[dIrat>0] = I0_nontransition[dIrat>0]
        I0[dIrat<0] = I0_transition[dIrat<0]

        Pj = I0 * R_sh * (Ib - I0)
        R0 = Pj / I0**2
        R0[I0 == 0] = 0

        return R0, I0, Pj

    def _compute_dc_params(self, R0_thresh=30e-3):
        """
        Calculates Ibias, dIbias, and dItes from axis manager, and then
        runs the DC param calc to estimate R0, I0, Pj, etc. Here you must

        Args:
            transition: (tuple)
                Range of voltage bias values (in low-cur units) where the
                "in-transition" resistance calculation should be used. If True,
                or False, will use in-transition or normal calc for all
                channels. Will default to ``cfg.dev.exp['transition_range']`` or
                (1, 8) if that does not exist or if self._cfg is not set.
            R0_thresh: (float)
                Any channel with resistance greater than R0_thresh will be
                unassigned from its bias group under the assumption that it's
                crosstalk

        Saves:
            Ibias:
                Array of shape (nbgs) containing the DC bias current for each
                bias group
            Vbias:
                Array of shape (nbgs) containing the DC bias voltage for each
                bias group
            dIbias:
                Array of shape (nbgs) containing the step current for each
                bias group
            dVbias:
                Array of shape (nbgs) containing the step voltage for each
                bias group
        """
        nbgs = len(self.am.biases)

        Ibias = np.full(nbgs, np.nan)
        dIbias = np.full(nbgs, 0.0, dtype=float)

        # Compute Ibias and dIbias
        bias_line_resistance = self.meta['bias_line_resistance']
        high_low_current_ratio = self.meta['high_low_current_ratio']
        rtm_bit_to_volt = self.meta['rtm_bit_to_volt']
        amp_per_bit = 2 * rtm_bit_to_volt / bias_line_resistance
        if self.high_current_mode:
            amp_per_bit *= high_low_current_ratio
        for bg in range(nbgs):
            if len(self.edge_idxs[bg]) == 0:
                continue
            b0 = self.am.biases[bg, self.edge_idxs[bg][0] - 3]
            b1 = self.am.biases[bg, self.edge_idxs[bg][0] + 3]
            Ibias[bg] = b0 * amp_per_bit
            dIbias[bg] = (b1 - b0) * amp_per_bit

        # Compute dItes
        i0 = np.nanmean(self.mean_resp[:, :5], axis=1)
        i1 = np.nanmean(self.mean_resp[:, -10:], axis=1)
        dItes = i1 - i0

        self.Ibias = Ibias
        self.Vbias = Ibias * bias_line_resistance
        self.dIbias = dIbias
        self.dVbias = dIbias * bias_line_resistance
        self.dItes = dItes

        R0, I0, Pj = self._compute_R0_I0_Pj()

        Si = -1./(I0 * (R0 - self.meta['R_sh']))

        # If resistance is too high, most likely crosstalk so just reset
        # bg mapping and det params
        if R0_thresh is not None:
            m = np.abs(R0) > R0_thresh
            self.bgmap[m] = -1
            for arr in [R0, I0, Pj, Si]:
                arr[m] = np.nan

        self.R0 = R0
        self.I0 = I0
        self.Pj = Pj
        self.Si = Si

        return R0, I0, Pj, Si

    def _fit_tau_effs(self, tmin=1.5e-3, weight_exp=0.3):
        """
        Fits mean step responses to exponential

        Args:
            tmin: float
                DEPRECATED!! do not use.

        Saves:
            step_fit_tmin: np.ndarray
                Array of shape (nchans) containing tmin used for each chan.
            step_fit_popts: np.ndarray
                Array of shape (nchans, 3) containing popt for each chan.
            step_fit_pcovs: np.ndarray
                Array of shape (nchans, 3, 3) containing pcov matrices for each
                channel
            tau_eff: np.ndarray
                Array of shape (nchans) contianing tau_eff for each chan.
        """
        nbgs = len(self.am.biases)
        nchans = len(self.am.signal)
        step_fit_popts = np.full((nchans, 3), np.nan)
        step_fit_pcovs = np.full((nchans, 3, 3), np.nan)
        step_fit_tmin = np.full(nchans, np.nan)

        for bg in range(nbgs):
            rcs = np.where(self.bgmap == bg)[0]
            if not len(rcs):
                continue
            ts = self.resp_times[bg]
            if not len(ts[~np.isnan(ts)]):
                continue
            for rc in rcs:
                resp = self.mean_resp[rc]
                tmin_m = np.abs(resp) > 0.9*np.nanmax(np.abs(resp))
                if not tmin_m.any():
                    continue
                tmin = np.max((0, ts[tmin_m][-1]))
                m = (ts > tmin) & (~np.isnan(resp))
                if not m.any():
                  continue
                offset_guess = np.nanmean(resp[np.abs(ts - ts[-1]) < 0.01])
                bounds = [
                    (-np.inf, 0, -np.inf),
                    (np.inf, 0.1, np.inf)
                ]
                p0 = (0.1, 0.001, offset_guess)
                try:
                    popt, pcov = scipy.optimize.curve_fit(
                        exp_fit, ts[m], resp[m],
                        sigma=ts[m]**weight_exp, p0=p0, bounds=bounds
                    )
                    step_fit_popts[rc] = popt
                    step_fit_pcovs[rc] = pcov
                    step_fit_tmin[rc] = tmin
                except RuntimeError:
                    pass

        self.step_fit_tmin = step_fit_tmin
        self.step_fit_popts = step_fit_popts
        self.step_fit_pcovs = step_fit_pcovs
        self.tau_eff = step_fit_popts[:, 1]

####################################################################
# Plotting functions
####################################################################

def plot_steps(bsa, rc, nsteps=5, offset=0, ax=None, **kw):
    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = ax.figure

    bg = bsa.bgmap[rc]
    i0 = bsa.edge_idxs[bg][2]
    if 3 + nsteps < len(bsa.edge_idxs[bg]):
        i1 = bsa.edge_idxs[bg][3 + nsteps]
    else:
        i1 = bsa.edge_idxs[bg][-1]

    sl = slice(i0, i1)
    ts = bsa.am.timestamps[sl]
    ts = ts - ts[0]
    sig = bsa.am.signal[rc, sl]
    sig = sig - np.nanmean(sig) + offset
    ax.plot(ts, sig, **kw)

    return fig, ax


def plot_step_fit(bsa, rc, ax=None, plot_all_steps=True):
    """
    Plots the step response and fit of a given readout channel
    """
    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = ax.figure

    bg = bsa.bgmap[rc]
    ts = bsa.resp_times[bg]
    try:
        m = ts > bsa.step_fit_tmin[rc]
    except TypeError:
        m = ts > bsa.step_fit_tmin
    if plot_all_steps:
        for sig in bsa.step_resp[rc]:
            ax.plot(ts*1000, sig, alpha=0.1, color='grey')
    ax.plot(ts*1000, bsa.mean_resp[rc], '.', label='avg step')
    ax.plot(ts[m]*1000, exp_fit(ts[m], *bsa.step_fit_popts[rc]), label='fit')

    text = '\n'.join([
        r'$\tau_\mathrm{eff}$=' + f'{bsa.tau_eff[rc]*1000:0.2f} ms',
        r'$R_\mathrm{TES}=$' + f'{bsa.R0[rc]* 1000:0.2f}' + r'm$\Omega$',
    ])

    ax.text(0.7, 0.1, text, transform=ax.transAxes,
            bbox={'facecolor': 'wheat', 'alpha': 0.3}, fontsize=12)

    ax.legend()
    ax.set(xlabel="Time (ms)", ylabel="Current (Amps)")

    return fig, ax

def plot_iv_res_comparison(bsa, lim=None, bgs=None, ax=None):
    if bgs is None:
        bgs = bsa.bgs
    bgs = np.atleast_1d(bgs)

    iva = iv.IVAnalysis.load(bsa.meta['iv_file'])
    chmap = sdl.map_band_chans(bsa.bands, bsa.channels,
                               iva.bands, iva.channels)
    iv_res = np.full(bsa.nchans, np.nan)

    if ax is None:
        fig, ax = plt.subplots()
        fig.patch.set_facecolor('white')
    else:
        fig = ax.figure

    for bg in bgs:
        m = bsa.bgmap == bg
        vb = bsa.Vbias[bg]
        idx = np.nanargmin(np.abs(iva.v_bias - vb))
        iv_res = iva.R[chmap[m], idx]
        ax.scatter(bsa.R0[m]*1000, iv_res*1000, marker='.', alpha=0.2,
                   label=f'Bias Group {bg}')
    if lim is None:
        lim = (-0.1, 10)
    ax.plot(lim, lim, ls='--', color='grey', alpha=0.4)
    ax.set(xlim=lim, ylim=lim)
    ax.set_xlabel("Bias Step Resistance (mOhm)")
    ax.set_ylabel("IV Resistance (mOhm)")

    return fig, ax

def get_plot_text(bsa):
    """
    Gets text to add to plot textboxes
    """
    return '\n'.join([
        f"stream_id: {bsa.meta['stream_id']}",
        f"sid: {bsa.sid}",
        f"path: {os.path.basename(bsa.filepath)}",
    ])

def plot_bg_assignment(bsa, text_loc=(0.05, 0.85), text_kw=None):
    """
    Plots bias group assignment summary
    """
    xs = np.arange(13)
    ys = np.zeros(13)

    xticklabels = []
    for i in range(12):
        ys[i] = np.sum(bsa.bgmap == i)
        xticklabels.append(str(i))
    ys[12] = np.sum(bsa.bgmap == -1)
    xticklabels.append('None')

    fig, ax = plt.subplots(figsize=(16, 10))
    ax.bar(xs, ys)
    ax.set_xticks(np.arange(13))
    ax.set_xticklabels(xticklabels)
    ax.set_xlabel("Bias Group", fontsize=16)
    ax.set_ylabel("Num Channels", fontsize=16)
    if text_kw is None:
        text_kw = {}
    
    ax.text(
        *text_loc, get_plot_text(bsa), transform=ax.transAxes,
        fontsize=18, bbox=dict(facecolor='white', alpha=0.8)
    )
    return fig, ax


def plot_Rtes(bsa):
    """
    Plots Rtes
    """
    fig, ax = plt.subplots()
    ax.hist(bsa.R0 * 1000, range=(-1, 10), bins=40)
    ax.set_xlabel(r"$R_\mathrm{TES}$ (m$\Omega$)", fontsize=12)
    return fig, ax

def plot_Rfrac(bsa, text_loc=(0.6, 0.8)):
    """
    Plots Rfrac split out by bias group
    """
    iva = iv.IVAnalysis.load(bsa.meta['iv_file'])
    chmap = sdl.map_band_chans(bsa.bands, bsa.channels,
                               iva.bands, iva.channels)

    Rfracs = bsa.R0 / iva.R_n[chmap]
    lim = (-0.1, 1.2)
    nbins=30
    bins = np.linspace(*lim, nbins)
    bgs = np.arange(12)
    hists = np.zeros((len(bgs), nbins-1))
    for bg in bgs:
        m = bsa.bgmap == bg
        if not m.any():
            continue
        hists[bg, :] = np.histogram(Rfracs[m], bins=bins)[0]

    offset = 0.2* np.max(hists)
    fig, ax = plt.subplots(figsize=(18, 10))
    fig.patch.set_facecolor('white')
    for bg, h in enumerate(hists):
        ax.plot(bins[1:], h + offset  * bg, '-o', markersize=4)
        ax.fill_between(bins[1:], h + offset * bg, y2=offset * bg, alpha=0.2)
        txt = f"{int(np.sum(h))} Chans"
        ax.text(1.1, offset*(bg + 0.15), txt, fontsize=20)

    ax.set_xlabel(r"Rfrac", fontsize=24)
    ax.set_yticks(offset * bgs)
    ax.set_yticklabels([f'BG {bg}' for bg in bgs], fontsize=20)
    ax.tick_params(axis='x', labelsize=24)
    txt = get_plot_text(bsa)
    ax.text(*text_loc, txt, bbox=dict(facecolor='white', alpha=0.8), fontsize=20,
            transform=ax.transAxes)
    return fig, ax


@dataclass
class BiasStepConfig:
    bgs: Optional[float] = None
    dc_voltage: float = None
    step_voltage: float = 0.05
    step_duration: float = 0.05
    nsteps: int = 20
    high_current_mode: bool = True
    hcm_wait_time: float = 3.
    dacs: str = 'pos'
    use_waveform: bool = True
    plot_rfrac: bool = True
    show_plots: bool = True
    run_analysis: bool = True
    channel_mask: Optional[np.ndarray] = None
    g3_tag: Optional[str] = None
    stream_subtype: str = 'bias_steps'
    enable_compression: bool = False
    analysis_kwargs: Optional[dict] = None

    def __post_init__(self):
        if self.analysis_kwargs is None:
            self.analysis_kwargs = {}

        if self.dacs not in ['pos', 'neg', 'both']:
            raise ValueError(f'dacs={self.dacs} not in ["pos", "neg", "both"]')

@dataclass
class BgMapConfig(BiasStepConfig):
    dc_voltage: float = 0.3
    step_voltage: float = 0.01
    hcm_wait_time: float = 0
    plot_rfrac: bool = False
    stream_subtype: str = 'bgmap'

    def __post_init__(self):
        super().__post_init__()

        # Makes sure these kwargs are set by default unless otherwise
        # specified
        kw = {
            'assignment_thresh': 0.3, 'create_bg_map': True,
            'save_bg_map': True
        }
        kw.update(self.analysis_kwargs)
        self.analysis_kwargs = kw

    def get_bias_step_cfg(self):
        return BiasStepConfig(
            **{f.name: getattr(self, f.name) for f in fields(BiasStepConfig)}
        )


@sdl.set_action()
def take_bgmap(S, cfg, **bgmap_pars):
    """
    Function to easily create a bgmap. This will set all bias group voltages
    to 0 (since this is best for generating the bg map), and run bias-steps
    with default parameters optimal for creating a bgmap.

    See also:
    ---------
    take_bias_steps
    """
    kw = deepcopy(cfg.dev.exp['bgmap_defaults'])
    kw.update(bgmap_pars)
    bgmap_cfg = BgMapConfig(**kw)

    if bgmap_cfg.bgs is None:
        bgmap_cfg.bgs = cfg.dev.exp['active_bgs']
    bgmap_cfg.bgs = np.atleast_1d(bgmap_cfg.bgs)

    bsa = take_bias_steps(S, cfg, **asdict(bgmap_cfg))

    if hasattr(bsa, 'bgmap'):
        fig, ax = plot_bg_assignment(bsa)
        sdl.save_fig(S, fig, 'bg_assignments.png')
        if bgmap_cfg.show_plots:
            plt.show()
        else:
            plt.close()

    for bg in bgmap_cfg.bgs:
        S.set_tes_bias_bipolar(bg, 0.)

    return bsa


@sdl.set_action()
def take_bias_steps(S, cfg, **bscfg_pars):
    """
    Takes bias step data at the specified DC voltage. Assumes bias lines
    are already in low-current mode (if they are in high-current this will
    not run correction). This function runs bias steps and returns a
    BiasStepAnalysis object, which can be used to easily view and re-analyze
    data.

    This function will first run a "bias group sweep", running multiple steps
    on each bias-line one at a time. This data is used to generate a bgmap.
    After, <nsteps> bias steps are played on all channels simultaneously.

    Parameters:
        S (SmurfControl):
            Pysmurf control instance
        cfg (DetConfig):
            Detconfig instance
        bgs ( int, list, optional):
            Bias groups to run steps on.
        dc_voltage : float, optional
            DC bias used at low end of bias step to avoid divide by zeros in
            the analysis. If unspecified, uses the current value.
        step_voltage (float):
            Step voltage in Low-current-mode units. (i.e. this will be divided
            by the high-low-ratio before running the steps in high-current
            mode)
        step_duration (float):
            Duration in seconds of each step
        nsteps (int):
            Number of steps to run
        high_current_mode (bool):
            If true, switches to high-current-mode. If False, leaves in LCM
            which runs through the bias-line filter, so make sure you
            extend the step duration to be like >2 sec or something
        hcm_wait_time (float):
            Time to wait after switching to high-current-mode.
        dacs : {'pos', 'neg', 'both'}
            Which group of DACs to play bias-steps on.
        run_analysis (bool):
            If True, will attempt to run the analysis to calculate DC params
            and tau_eff. If this fails, the analysis object will
            still be returned but will not contain all analysis results.
        analysis_kwargs (dict, optional):
            Keyword arguments to be passed to the BiasStepAnalysis run_analysis
            function.
        channel_mask : np.ndarray, optional
            Mask containing absolute smurf-channels to write to disk
        g3_tag: string, optional
            Tag to attach to g3 stream.
        stream_subtype : optional, string
            Stream subtype for this operation. This will default to 'bias_steps'.
        enable_compression: bool, optional
            If True, will tell the smurf-streamer to compress G3Frames. Defaults
            to False because this dominates frame-processing time for high
            data-rate streams.
        plot_rfrac : bool
            Create rfrac plot, publish it, and save it. Default is True.
        show_plots : bool
            Show plot in addition to saving when running interactively. Default is False.
    """
    kw = deepcopy(cfg.dev.exp['biasstep_defaults'])
    kw.update(bscfg_pars)
    bscfg = BiasStepConfig(**kw)

    if bscfg.bgs is None:
        bscfg.bgs = cfg.dev.exp['active_bgs']
    bscfg.bgs = np.atleast_1d(bscfg.bgs)

    # Adds to account for steps that may be cut in analysis
    bscfg.nsteps += 4

    if bscfg.dc_voltage is not None:
        for bg in bscfg.bgs:
            S.set_tes_bias_bipolar(bg, bscfg.dc_voltage)

    initial_ds_factor = S.get_downsample_factor()
    initial_filter_disable = S.get_filter_disable()
    initial_dc_biases = S.get_tes_bias_bipolar_array()
    init_current_mode = sdl.get_current_mode_array(S)

    try:
        dc_biases = initial_dc_biases
        if bscfg.high_current_mode:
            dc_biases = dc_biases / S.high_low_current_ratio
            bscfg.step_voltage /= S.high_low_current_ratio
            sdl.set_current_mode(S, bscfg.bgs, 1)
            S.log(f"Waiting {bscfg.hcm_wait_time} sec after switching to hcm")
            time.sleep(bscfg.hcm_wait_time)

        bsa = BiasStepAnalysis(S, cfg, run_kwargs=asdict(bscfg))

        bsa.sid = sdl.stream_g3_on(
            S, tag=bscfg.g3_tag, channel_mask=bscfg.channel_mask, downsample_factor=1,
            filter_disable=True, subtype=bscfg.stream_subtype,
            enable_compression=bscfg.enable_compression
        )

        bsa.start = time.time()

        for bg in bscfg.bgs:
            if bscfg.use_waveform:
                play_bias_steps_waveform(
                    S, cfg, bg, bscfg.step_duration, bscfg.step_voltage,
                    num_steps=bscfg.nsteps
                )
            else:
                play_bias_steps_dc(
                    S, cfg, bg, bscfg.step_duration, bscfg.step_voltage,
                    num_steps=bscfg.nsteps, dacs=bscfg.dacs,
                )
        bsa.stop = time.time()

    finally:
        sdl.stream_g3_off(S)

        # Restores current mode to initial values
        sdl.set_current_mode(S, np.where(init_current_mode == 0)[0], 0)
        sdl.set_current_mode(S, np.where(init_current_mode == 1)[0], 1)

        S.set_downsample_factor(initial_ds_factor)
        S.set_filter_disable(initial_filter_disable)

    if bscfg.run_analysis:
        S.log("Running bias step analysis")
        try:
            bsa.run_analysis(save=True, **bscfg.analysis_kwargs)
            if bscfg.plot_rfrac:
                fig, ax = plot_Rfrac(bsa)
                path = sdl.make_filename(S, 'bs_rfrac_summary.png', plot=True)
                fig.savefig(path)
                S.pub.register_file(path, 'bs_rfrac_summary', plot=True, format='png')
                if not bscfg.show_plots:
                    plt.close(fig)
        except Exception:
            print(f"Bias step analysis failed with exception:")
            print(traceback.format_exc())

    return bsa


def plot_taueff_hist(bsa, text_loc=(0.4, 0.8), fontsize=15, figsize=(10, 6),
                     range=(0, 3)):
    """
    Plots a histogram of the tau_eff measurements for all bias lines.

    Args
    -----
    bsa : BiasStepAnalysis
        analyzed BiasStepAnalysis object
    text_loc : tuple[float]
        Location of textbox in the axis coordinate system.
    fontsize : int
        font size of text in textbox
    figsize : tuple
        Figure size
    range : tuple
        histogram range
    """
    fig, ax = plt.subplots(figsize=figsize)
    ax.hist(bsa.tau_eff * 1000, range=range, bins=30)
    ax.set_xlabel("Tau eff (ms)", fontsize=18)
    txt = get_plot_text(bsa)
    ax.text(*text_loc, txt, bbox=dict(facecolor='white', alpha=0.8),
            fontsize=fontsize, transform=ax.transAxes)
    return fig, ax



import time
import os
import traceback
import numpy as np
import sodetlib as sdl
import matplotlib.pyplot as plt
import scipy.optimize
from scipy.signal import welch
from sodetlib.operations import bias_steps, iv

np.seterr(all='ignore')

def play_bias_wave(S, cfg, bias_group, freqs_wave, amp_wave, duration,
                   dc_bias=None):
    """
    Play a sine wave on the bias group.

    Args
    ----
    bias_group : int
        The bias group
    freq_wave : float array
        List of frequencies to play. Unit = Hz.
    amp_wave : float
        Amplitude of sine wave to use. Unit = Volts.
    duration : float
        Duration each sine wave is played  
    dc_bias : float, optional
        Offset voltage of sine wave. Unit = Volts.

    Returns
    -------
    start_times : float list
        unix timestamp for the beginning of each sine wave.
    stop_times : float list
        unix timestamp for the end of each sine wave.
    """
    start_times, stop_times = [], []

    if dc_bias is None:
        dc_bias = S.get_tes_bias_bipolar_array()[bias_group]

    for freq in freqs_wave:
        S.log(f"BL sine wave with bg={bias_group}, freq={freq}")
        S.play_sine_tes(bias_group=bias_group,
                        tone_amp=amp_wave,
                        tone_freq=freq, dc_amp=dc_bias)
        start_times.append(time.time())
        time.sleep(duration)
        stop_times.append(time.time())
        S.set_rtm_arb_waveform_enable(0)
        S.set_tes_bias_bipolar(bias_group, dc_bias)
    return start_times, stop_times

def get_amplitudes(f_c, x, fs = 4000, N = 12000, window = 'hann'):
    """
    Function for calculating the amplitude of a sine wave.

    Args
    ----
    f_c : float
        Target frequency to calculate sine wave amplitude at.
    x : np.array
        Data to analyze. Either can be shape nsamp or ndet x nsamp.
    fs : int
        Sample rate. Unit = samples/second.
    N : int
        Number of samples to calculate FFT on.
    window : str
        Window function to use. See scipy.signal.get_window for a list of 
        valid window functions to use. Default is ``hann``.

    Returns
    -------
    a_peak : float
        Amplitudes of sine waves. Shape is len(x)
    """
    x = np.atleast_2d(x)
    f, p = welch(x, fs = fs, nperseg = N, 
                        scaling = 'spectrum',return_onesided=True,
                        window = window)
    a = np.sqrt(p)
    # Ran into problem with f == f_c when nperseg and len(x) are not right.
    # Ben to add a function to check this or enforce correct choice somehow.
    idx = np.argmin(np.abs(f-f_c))
    a_rms = a[:, idx]
    a_peak = np.sqrt(2)*a_rms
    return a_peak

def get_amplitudes_deprojection(f_c, x, ts):
    """
    Function for calculating the amplitude and phase of a wave by deprojecting sine and cosine components of desired frequency.

    Args
    ----
    f_c : float
        Target frequency to calculate wave amplitude at.
    x : np.array
        Data to analyze. Either can be shape nsamp or ndet x nsamp.
    ts : np.array
        Timestamps of data to analyze. Shape is nsamp.

    Returns
    -------
    a_peak : float
        Amplitudes of sine wave of the desired frequency. Shape is len(x)
    phase : float
        Phase of sine wave of the desired frequency. Shape is len(x)
    """

    vects = np.zeros((2, len(ts)), dtype='float32')
    vects[0, :] = np.sin(2*np.pi*f_c*ts)
    vects[1, :] = np.cos(2*np.pi*f_c*ts)
    I = np.linalg.inv(np.tensordot(vects, vects, (1, 1)))
    coeffs = np.matmul(x, vects.T)
    coeffs = np.dot(I, coeffs.T).T
    coeffs = np.atleast_2d(coeffs)

    a_peak = np.sqrt(coeffs[:,0]**2 + coeffs[:,1]**2)
    phase = np.arctan2(coeffs[:,0], coeffs[:,1])

    return a_peak, phase

class BiasWaveAnalysis:
    """
    Container to manage analysis of bias waves taken with the take_bias_waves
    function. The main function is ``run_analysis`` and will do a series of
    analysis procedures to calculate DC detector parameters from a single bias 
    wave frequency. The method will be extended to calculate tau_eff from multiple
    frequencies. Procedure:

     - Loads an axis manager with all the data
     - Finds locations of bias wave frequencies for each bias group
     - Gets detector response to each bias wave
     - Computes DC params R0, I0, Pj, Si from step responses
     - Later: fit multiple TES amplitudes with one-pole filter for tau_eff measurement.

    Most analysis inputs and products will be saved to a npy file so they can
    be loaded and re-analyzed easily on another computer like simons1.

    To load data from an saved step file, you can run::

        bwa = BiasWaveAnalysis.load(<path>)

    Attributes:
        bands (array (int) of shape (ndets)):
            Smurf bands for each resonator.
        channels (array (int) of shape (ndets)): 
            Smurf channel for each resonator.
        meta (dict) :
            Dictionary of smurf metadata.
        sid (int) :
            Session-id of streaming session
        run_kwargs :
            input kwargs
        start, stop (float):
            start and stop time of all steps
        high_current_mode (Bool) :
            If high-current-mode was used
        start_times, stop_times (array (float) of shape (nbgs, nfreqs)) :
            Arrays of start and stop times for the start of each frequency sine wave.
        bgmap (array (int) of shape (nchans)):
            Map from readout channel to assigned bias group. -1 means not
            assigned (that the assignment threshold was not met for any of the
            12 bgs)
        polarity (array (int) of shape (ndets)):
            The sign of each channel.
        resp_times (array (float) shape (nbgs, nfreqs, npts)):
            Shared timestamps for each of the wave responses in <wave_resp> and
            <mean_resp> with respect to the bg-wave location, with the wave
            occuring at t=0.
        mean_resp (array (float) shape (nchans, nfreqs, npts)):
            Wave response averaged accross all bias steps for a given channel
            in Amps.
        wave_resp (array (float) shape (nchans, nfreqs, npts)):
            Each individual wave response for a given channel in amps.
        wave_biases (array (float) shape (nbgs, nfreqs, npts)):
            Each individual bias wave for a given channel.
        Ibias (array (float) shape (nbgs)):
            DC bias current of each bias group (amps). DC Params are 
            only calculated off of the minimum frequency in the array of frequencies.
        Vbias:
            DC bias voltage of each bias group (volts in low-current mode)
        dIbias (array (float) shape (nbgs)):
            Wave current amplitude for each bias group (amps)
        dVbias (array (float) shape (nbgs)):
            Wave voltage amplitude for each bias group (volts in low-current mode)
        dItes (array (float) shape (nchans)):
            Array of tes wave response height for each channel (amps)
        R0 (array (float) shape (nchans)):
            Computed TES resistances for each channel (ohms). 
        I0 (array (float) shape (nchans)):
            Computed TES currents for each channel (amps)
        Pj (array (float) shape (nchans)):
            Bias power computed for each channel
        Si (array (float) shape (nchans)):
            Responsivity computed for each channel
        R_n_IV (array (float) shape (nchans)):
            Array of normal resistances for each channel pulled from IV
            in the device cfg.
        Rfrac (array (float) shape (nchans)):
            Rfrac of each channel, determined from R0 and the channel's normal
            resistance.
    """
    def __init__(self, S=None, cfg=None, bgs=None, run_kwargs=None):
        self._S = S
        self._cfg = cfg

        self.bgs = bgs
        self.am = None

        if S is not None:
            self.meta = sdl.get_metadata(S, cfg)
            self.stream_id = cfg.stream_id

            if run_kwargs is None:
                run_kwargs = {}
            self.run_kwargs = run_kwargs
            self.high_current_mode = run_kwargs.get("high_current_mode", True)
          
    def save(self, path=None):
        data = {}
        saved_fields = [
            # Run data and metadata
            'bands', 'channels', 'sid', 'meta', 'run_kwargs', 'start', 'stop',
            'high_current_mode', 'start_times', 'stop_times',
            # Bgmap data
            'bgmap', 'polarity',
            # Step data and fits, including chunked bias data
            'resp_times', 'mean_resp', 'wave_resp', 'wave_biases',
            # Add in tau fit stuff here. The below commented out params are anticipated to be reported for tau analysis. 
            # 'step_fit_tmin', 'step_fit_popts', 'step_fit_pcovs',
            # 'tau_eff',
            # Det param data
            'Ibias', 'Vbias', 'dIbias', 'dVbias', 'dItes',
            'R0', 'I0', 'Pj', 'Si',
            # From IV's
            'R_n_IV', 'Rfrac',
        ]

        for f in saved_fields:
            if not hasattr(self, f):
                print(f"WARNING: field {f} does not exist... "
                      "defaulting to None")
            data[f] = getattr(self, f, None)

        if path is not None:
            np.save(path, data, allow_pickle=True)
            self.filepath = path
        else:
            self.filepath = sdl.validate_and_save(
                'bias_wave_analysis.npy', data, S=self._S, cfg=self._cfg,
                make_path=True
            )

    @classmethod
    def load(cls, filepath):
        self = cls()
        data = np.load(filepath, allow_pickle=True).item()
        for k, v in data.items():
            setattr(self, k, v)
        self.filepath = filepath
        return self
    
    def run_analysis(self, arc=None, base_dir='/data/so/timestreams',
                     R0_thresh=30e-3, save=False, bg_map_file=None):
        """
        Analyzes data taken with take_bias_waves.

        Parameters
        ----------
        arc (optional, G3tSmurf):
                G3tSmurf archive. If specified, will attempt to load
                axis-manager using archive instead of sid.
        base_dir (optiional, str):
                Base directory where timestreams are stored. Defaults to
                /data/so/timestreams.
        R0_thresh (float):
                Any channel with resistance greater than R0_thresh will be
                unassigned from its bias group under the assumption that it's
                crosstalk
        save (bool):
                If true will save the analysis to a npy file.
        bg_map_file (optional, path):
                If create_bg_map is false and this file is not None, use this file
                to load the bg_map.

        CURRENTLY ONLY INCLUDES CALCULATION OF DC PARAMETERS.
        NO TIME CONSTANT FITS.
        """
        self._load_am(arc=arc, base_dir=base_dir)
        self._split_frequencies()
        if bg_map_file is not None:
            self.bgmap, self.polarity = sdl.load_bgmap(
                self.bands, self.channels, bg_map_file)
        else:
            self.bgmap, self.polarity = sdl.load_bgmap(
                self.bands, self.channels, self.meta['bgmap_file'])

        # GOT TO HERE
        self._get_wave_response()
        self._compute_dc_params(R0_thresh=R0_thresh)

        # Load R_n from IV
        self.R_n_IV = np.full(self.nchans, np.nan)
        # Rfrac determined from R0 and R_n
        self.Rfrac = np.full(self.nchans, np.nan)
        if self.meta['iv_file'] is not None:
            if os.path.exists(self.meta['iv_file']):
                iva = iv.IVAnalysis.load(self.meta['iv_file'])
                chmap = sdl.map_band_chans(
                    self.bands, self.channels, iva.bands, iva.channels
                )
                self.R_n_IV = iva.R_n[chmap]
                self.R_n_IV[chmap == -1] = np.nan
                self.Rfrac = self.R0 / self.R_n_IV

        # Can add in function for fitting time constant from multifrequency data here.
        # self._fit_tau_effs(tmin=fit_tmin)

        if save:
            self.save()

    def _load_am(self, arc=None, base_dir='/data/so/timestreams', fix_timestamps=True):
        """
        Attempts to load the axis manager from the sid or return one that's
        already loaded. Also sets the `abs_chans` array.
        """
        if self.am is None:
            if arc:
                self.am = arc.load_data(self.start, self.stop, stream_id=self.meta['stream_id'])
            else:
                self.am = sdl.load_session(self.meta['stream_id'], self.sid,
                                            base_dir=base_dir)

            # Fix up timestamp jitter from timestamping in software
            if fix_timestamps:
                fsamp, t0 = np.polyfit(self.am.primary['FrameCounter'],
                                        self.am.timestamps, 1)
                self.am.timestamps = t0 + self.am.primary['FrameCounter']*fsamp
            if "det_info" in self.am:
                self.bands = self.am.det_info.smurf.band
                self.channels = self.am.det_info.smurf.channel
            else:
                self.bands = self.am.ch_info.band
                self.channels = self.am.ch_info.channel
            self.abs_chans = self.bands*512 + self.channels
            self.nbgs = len(self.am.biases)
            self.nchans = len(self.am.signal)
        return self.am

    def _split_frequencies(self, am=None):
        """
        Gets indices for each sine wave frequency on each bias group.

        Returns
        -------
        start_idxs : float array
            Array of indices for the start of each frequency sine wave.
            Shape (n_bias_groups, n_frequencies).
        stop_idxs : float array
            Array of indices for the end of each frequency sine wave.
            Shape (n_bias_groups, n_frequencies).
        """
        if am is None:
            am = self.am

        self.start_idxs = np.full(np.shape(self.start_times), -1)
        self.stop_idxs = np.full(np.shape(self.stop_times), -1)

        for bg, bias in enumerate(am.biases[:12]):
            if np.all(np.isnan(self.start_times[bg,:])):
                continue
            for i, [start, stop] in enumerate(zip(self.start_times[bg,:], self.stop_times[bg,:])):
                self.start_idxs[bg, i] = np.argmin(np.abs(am.timestamps - start))
                self.stop_idxs[bg, i] = np.argmin(np.abs(am.timestamps - stop))

        return self.start_idxs, self.stop_idxs
    
    def _get_wave_response(self, am=None):
        """
        Splits up full axis manager into each sine wave.
        Here we enforce the number of points in each sine wave to be equal.

        Returns
        -------
        ts : float array
            Array of timestamps for each sine wave period
            Shape (n_bias_groups, n_frequencies, n_pts_in_sine_wave).
        sigs : float array
            Array of sine response data.
            Shape (n_detectors, n_frequencies, n_pts_in_sine_wave).
        UPDATE THE DOCSTRING
        """
        if am is None:
            am = self.am

        nchans = len(am.signal)
        nbgs = 12
        n_freqs = np.shape(self.start_idxs)[-1]
        npts = np.nanmin(self.stop_idxs-self.start_idxs)

        sigs = np.full((nchans, n_freqs, npts), np.nan)
        biases = np.full((nbgs, n_freqs, npts), np.nan)
        ts = np.full((nbgs, n_freqs, npts), np.nan)

        A_per_rad = self.meta['pA_per_phi0'] / (2*np.pi) * 1e-12
        for bg in np.unique(self.bgmap):
            if bg == -1:
                continue
            rcs = np.where(self.bgmap == bg)[0]
            for i, si in enumerate(self.start_idxs[bg]):
                if np.isnan(ts[bg,i]).all():
                    ts[bg, i, :] = am.timestamps[si:si+npts] - am.timestamps[si]
                sigs[rcs, i, :] = am.signal[rcs, si:si+npts] * A_per_rad
                biases[bg, i, :] = am.biases[bg, si:si+npts]

        self.resp_times = ts
        self.wave_resp = (sigs.T * self.polarity).T
        self.mean_resp = (np.nanmean(sigs, axis=1).T * self.polarity).T
        self.wave_biases = biases

        return ts, sigs, biases
    
    def _compute_dc_params(self, R0_thresh=30e-3):
        """
        Calculates Ibias, dIbias, and dItes from axis manager, and then
        runs the DC param calc to estimate R0, I0, Pj, etc. 
        If multiple frequency are taken, the DC Params are only calculated
        off of the minimum frequency in the array of frequencies.

        Args:
            R0_thresh: (float)
                Any channel with resistance greater than R0_thresh will be
                unassigned from its bias group under the assumption that it's
                crosstalk

        Saves:
            Ibias:
                Array of shape (nbgs) containing the DC bias current for each
                bias group
            Vbias:
                Array of shape (nbgs) containing the DC bias voltage for each
                bias group
            dIbias:
                Array of shape (nbgs) containing the wave current amplitude for
                each bias group
            dVbias:
                Array of shape (nbgs) containing the wave voltage amplitude for
                each bias group
            dItes:
                Array of shape (nchans) containing the wave current amplitude for
                each detector.
        """
        nbgs = 12
        nchans = len(self.am.signal)
        npts = np.nanmin(self.stop_idxs-self.start_idxs)

        Ibias = np.full(nbgs, np.nan)
        dIbias = np.full(nbgs, 0.0, dtype=float)
        dItes = np.full(nchans, np.nan)

        dIbias_phase = np.full(nbgs, 0.0, dtype=float)
        dItes_phase = np.full(nchans, np.nan)

        # Compute Ibias and dIbias
        bias_line_resistance = self.meta['bias_line_resistance']
        high_low_current_ratio = self.meta['high_low_current_ratio']
        rtm_bit_to_volt = self.meta['rtm_bit_to_volt']
        amp_per_bit = 2 * rtm_bit_to_volt / bias_line_resistance
        if self.high_current_mode:
            amp_per_bit *= high_low_current_ratio
        
        for bg in range(nbgs):
            if len(self.start_idxs[bg]) == 0:
                continue
            rcs = np.where(self.bgmap == bg)[0]
            s = slice(int(self.start_idxs[bg][0]),
                      int(self.start_idxs[bg][0] + npts))
            Ibias[bg] = np.nanmean(self.am.biases[bg, s]) * amp_per_bit
            
            dIbias[bg], dIbias_phase[bg] = get_amplitudes_deprojection(self.run_kwargs['freqs_wave'][0],
                                        self.am.biases[bg, s], self.resp_times[bg, 0, :])
            dIbias[bg] = dIbias[bg] * amp_per_bit
          
            dItes[rcs], dItes_phase[rcs] = get_amplitudes_deprojection(self.run_kwargs['freqs_wave'][0],
                                        self.wave_resp[rcs,0,:], self.resp_times[bg, 0, :])

        self.Ibias = Ibias
        self.Vbias = Ibias * bias_line_resistance
        self.dIbias = dIbias
        self.dIbias_phase = dIbias_phase
        self.dVbias = dIbias * bias_line_resistance
        self.dItes = dItes
        self.dItes_phase = dItes_phase

        R0, I0, Pj = self._compute_R0_I0_Pj()

        Si = -1./(I0 * (R0 - self.meta['R_sh']))

        # If resistance is too high, most likely crosstalk so just reset
        # bg mapping and det params
        if R0_thresh is not None:
            m = np.abs(R0) > R0_thresh
            self.bgmap[m] = -1
            for arr in [R0, I0, Pj, Si]:
                arr[m] = np.nan

        self.R0 = R0
        self.I0 = I0
        self.Pj = Pj
        self.Si = Si

        return R0, I0, Pj, Si

    def _compute_R0_I0_Pj(self):
        """
        Computes the DC params R0 I0 and Pj

        Carbon copy from BiasStepAnalysis
        """
        Ib = self.Ibias[self.bgmap]
        dIb = self.dIbias[self.bgmap]
        dItes = self.dItes

        dIb_phase = self.dIbias_phase[self.bgmap]
        dItes_phase = self.dItes_phase
      
        Ib[self.bgmap == -1] = np.nan
        dIb[self.bgmap == -1] = np.nan

        #sign of phase response, is there a better way to do it?
        rel_phase = dItes_phase - dIb_phase 
        rel_phase  = np.isclose(abs(rel_phase), np.pi, atol=3e-1, rtol = 1e-1) #arbitrary cutoff?
        rel_phase_sign = np.full(rel_phase.shape, 1.0)
        rel_phase_sign[np.where(rel_phase == True)] = -1.0
      
        dIrat = rel_phase_sign * (dItes / dIb)
    
        R_sh = self.meta["R_sh"]

        I0 = np.zeros_like(dIrat)
        I0_nontransition = Ib * dIrat
        I0_transition = Ib * dIrat / (2 * dIrat - 1)
        I0[dIrat>0] = I0_nontransition[dIrat>0]
        I0[dIrat<0] = I0_transition[dIrat<0]

        Pj = I0 * R_sh * (Ib - I0)
        R0 = Pj / I0**2
        R0[I0 == 0] = 0

        return R0, I0, Pj

@sdl.set_action()
def take_bias_waves(S, cfg, bgs=None, amp_wave=0.05, freqs_wave=[23.0],
                    duration=10, high_current_mode=True, hcm_wait_time=3,
                    run_analysis=True, analysis_kwargs=None, channel_mask=None,
                    g3_tag=None, stream_subtype='bias_waves',
                    enable_compression=False, plot_rfrac=True, show_plots=False):
    """
    Takes bias wave data at the current DC voltage. Assumes bias lines
    are already in low-current mode (if they are in high-current this will
    not run correction). This function runs bias waves and returns a
    BiasWaveAnalysis object, which can be used to easily view and re-analyze
    data.

    Parameters:
        S (SmurfControl):
            Pysmurf control instance
        cfg (DetConfig):
            Detconfig instance
        bgs ( int, list, optional):
            Bias groups to run steps on, defaulting to all 12. Note that the
            bias-group mapping generated by the bias step analysis will be
            restricted to the bgs set here so if you only run with a small
            subset of bias groups, the map might not be correct.
        amp_wave (float):
            Bias wave amplitude voltage in Low-current-mode units.
            This will be divided by the high-low-ratio before running the steps
            in high-current mode.
        freqs_wave (float):
            List of frequencies to take bias wave data at.
        duration (float):
            Duration in seconds of bias wave frequency
        high_current_mode (bool):
            If true, switches to high-current-mode. If False, leaves in LCM
            which runs through the bias-line filter, so make sure you
            extend the step duration to be like >2 sec or something
        hcm_wait_time (float):
            Time to wait after switching to high-current-mode.
        run_analysis (bool):
            If True, will attempt to run the analysis to calculate DC params
            and tau_eff. If this fails, the analysis object will
            still be returned but will not contain all analysis results.
        analysis_kwargs (dict, optional):
            Keyword arguments to be passed to the BiasWaveAnalysis run_analysis
            function.
        channel_mask : np.ndarray, optional
            Mask containing absolute smurf-channels to write to disk
        g3_tag: string, optional
            Tag to attach to g3 stream.
        stream_subtype : optional, string
            Stream subtype for this operation. This will default to 'bias_waves'.
        enable_compression: bool, optional
            If True, will tell the smurf-streamer to compress G3Frames. Defaults
            to False because this dominates frame-processing time for high
            data-rate streams.
        plot_rfrac : bool
            Create rfrac plot, publish it, and save it. Default is True.
        show_plots : bool
            Show plot in addition to saving when running interactively. Default is False.
    """
    if bgs is None:
        bgs = cfg.dev.exp['active_bgs']
    bgs = np.atleast_1d(bgs)

    # Dumb way to get all run kwargs, but we probably want to save these in
    # data object
    freqs_wave = np.sort(np.atleast_1d(freqs_wave)) # Enforces lowest frequency first.
    run_kwargs = {
        'bgs': bgs, 'amp_wave': amp_wave,
        'freqs_wave': freqs_wave, 'duration': duration,
        'high_current_mode': high_current_mode,
        'hcm_wait_time': hcm_wait_time, 'run_analysis': run_analysis,
        'analysis_kwargs': analysis_kwargs, 'channel_mask': channel_mask,
    }

    initial_ds_factor = S.get_downsample_factor()
    initial_filter_disable = S.get_filter_disable()
    initial_dc_biases = S.get_tes_bias_bipolar_array()

    try:
        dc_biases = initial_dc_biases
        init_current_mode = sdl.get_current_mode_array(S)
        if high_current_mode:
            dc_biases = dc_biases / S.high_low_current_ratio
            amp_wave /= S.high_low_current_ratio
            sdl.set_current_mode(S, bgs, 1)
            S.log(f"Waiting {hcm_wait_time} sec after switching to hcm")
            time.sleep(hcm_wait_time)

        bwa = BiasWaveAnalysis(S, cfg, bgs, run_kwargs=run_kwargs)

        bwa.sid = sdl.stream_g3_on(
            S, tag=g3_tag, channel_mask=channel_mask, downsample_factor=1,
            filter_disable=True, subtype=stream_subtype, enable_compression=enable_compression
        )

        bwa.start_times = np.full((12, len(freqs_wave)), np.nan)
        bwa.stop_times = np.full((12, len(freqs_wave)), np.nan)

        bwa.start = time.time()

        for bg in bgs:
            bwa.start_times[bg,:], bwa.stop_times[bg,:] = play_bias_wave(S, cfg, bg,
                                                                        freqs_wave,
                                                                        amp_wave, duration)

        bwa.stop = time.time()

    finally:
        sdl.stream_g3_off(S)

        # Restores current mode to initial values
        sdl.set_current_mode(S, np.where(init_current_mode == 0)[0], 0)
        sdl.set_current_mode(S, np.where(init_current_mode == 1)[0], 1)

        S.set_downsample_factor(initial_ds_factor)
        S.set_filter_disable(initial_filter_disable)

    if run_analysis:
        S.log("Running bias wave analysis")
        try:
            if analysis_kwargs is None:
                analysis_kwargs = {}
            bwa.run_analysis(save=True, **analysis_kwargs)
            if plot_rfrac:
                fig, _ = bias_steps.plot_Rfrac(bwa)
                path = sdl.make_filename(S, 'bw_rfrac_summary.png', plot=True)
                fig.savefig(path)
                S.pub.register_file(path, 'bw_rfrac_summary', plot=True, format='png')
                if not show_plots:
                    plt.close(fig)
        except Exception:
            print(f"Bias wave analysis failed with exception:")
            print(traceback.format_exc())

    return bwa



import argparse
import os
import yaml
from collections import OrderedDict
import sys
import time
import shutil


class YamlReps:
    class FlowSeq(list):
        """Represents a list as a flow sequencey by default"""
        pass

    class Odict(OrderedDict):
        """
        Represents an ordered dict in the same way as a dict, but with
        ordering upheld.
        """
        pass

    @classmethod
    def setup_reps(cls):
        def flowseq_rep(dumper, data):
            return dumper.represent_sequence(u'tag:yaml.org,2002:seq', data,
                                             flow_style=True)
        yaml.add_representer(cls.FlowSeq, flowseq_rep)

        def odict_rep(dumper, data):
            return dumper.represent_mapping('tag:yaml.org,2002:map',data.items())
        yaml.add_representer(cls.Odict, odict_rep)


YamlReps.setup_reps()

# Default values for device cfg entries
exp_defaults = {
    # General stuff
    'downsample_factor': 20, 'coupling_mode': 'dc', 'synthesis_scale': 1,
    'active_bands': [0, 1, 2, 3, 4, 5, 6, 7],
    'active_bgs': [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11],
    'downsample_mode': 'internal',
    'enable_compression': True,

    # Downsample filter parameters
    # cutoff freq in Hz reverse engineered to produce pysmurf default filter
    "downsample_filter": {"order": 4, "cutoff_freq": 63.0},

    # Dict of device-specific default params to use for take_iv
    'iv_defaults': {},

    # Dict of device-specific default params to use for take_bgmap
    'bgmap_defaults': {},

    # Dict of device-specific default params to use for take_bias_steps
    'biasstep_defaults': {},

    # Amp stuff
    "amps_to_bias": ['hemt', 'hemt1', 'hemt2', '50k', '50k1', '50k2'],
    "amp_enable_wait_time": 10.0, "amp_step_wait_time": 0.2,

    "amp_50k_init_gate_volt": -0.5, "amp_50k_drain_current": 15.0,
    "amp_50k_gate_volt": None, "amp_50k_drain_current_tolerance": 0.2,
    "amp_hemt_init_gate_volt": -1.0, "amp_hemt_drain_current": 8.0,
    "amp_hemt_gate_volt": None, "amp_hemt_drain_current_tolerance": 0.2,

    "amp_50k1_init_gate_volt": -0.5, "amp_50k1_drain_current": 15.0,
    "amp_50k1_gate_volt": None, "amp_50k1_drain_current_tolerance": 0.2,
    "amp_50k1_drain_volt": 4,
    "amp_50k2_init_gate_volt": -0.5, "amp_50k2_drain_current": 15.0,
    "amp_50k2_gate_volt": None, "amp_50k2_drain_current_tolerance": 0.2,
    "amp_50k2_drain_volt": 4,

    "amp_hemt1_init_gate_volt": -0.8, "amp_hemt1_drain_current": 8.0,
    "amp_hemt1_gate_volt": None, "amp_hemt1_drain_current_tolerance": 0.2,
    "amp_hemt1_drain_volt": 0.6,
    "amp_hemt2_init_gate_volt": -0.8, "amp_hemt2_drain_current": 8.0,
    "amp_hemt2_gate_volt": None, "amp_hemt2_drain_current_tolerance": 0.2,
    "amp_hemt2_drain_volt": 0.6,

    # Find freq
    'res_amp_cut': 0.01, 'res_grad_cut': 0.01,

    # Tracking stuff
    "flux_ramp_rate_khz": 4, "init_frac_pp": 0.4, "nphi0": 5,
    "f_ptp_range": [10, 200], "df_ptp_range": [0, 100], "r2_min": 0.9,
    "min_good_tracking_frac": 0.8,
    'feedback_start_frac': 0.02, 'feedback_end_frac': 0.98,

    # Misc files
    "tunefile": None, "bgmap_file": None, "iv_file": None,
    "res_fit_file": None,

    # Overbiasing
    'overbias_wait': 2.0,
    'overbias_sleep_time_sec': 300,

    # Biasing
    'rfrac':[0.3, 0.6],
}
band_defaults = {
    # General
    "band_delay_us": None, "uc_att": None, "dc_att": None,
    "attens_optimized": False, "tone_power": 12,

    # Band-specific tracking stuff
    "lms_gain": 0, "feedback_gain": 2048, "frac_pp": None, "lms_freq_hz": None,

    # Gradient Descent Parameters
    "gradientDescentMaxIters": 100,
    "gradientDescentAverages": 2, 
    "gradientDescentGain": 0.001,
    "gradientDescentConvergeHz": 500,
    "gradientDescentStepHz": 5000,
    "gradientDescentMomentum": 1,
    "gradientDescentBeta": 0.1,

    # Fixed tones
    "fixed_tones": {"enabled": False, "freq_offsets": [], "channels": [], "tone_power": 0},
}
bg_defaults = {
    'overbias_voltage': 19.9,
    'cool_voltage': 10.0,
    'testbed_100mK_bias_voltage': 10.0
}


class DeviceConfig:
    """
    Configuration object containing all "device" specific information. That is,
    parameters that will probably change based on which test-bed you are
    using and what hardware is loaded into that test-bed. Device configuration
    info is split into three groups: ``experiment``, ``bias_groups`` and
    ``bands``.

    The ``experiment`` group contains data that is relevant to the whole
    experiment, e.g. the tunefile that should be loaded or the amplifier
    currents and voltages that should be set.

    The ``bias_groups`` group contains data for each individual bias group.
    For instance this is where the bias_high/low/step values are stored for each
    bias group.

    The ``bands`` group contains data for each individual band. For instance,
    the dc_att and drive values.

    Attributes:
        exp (dict):
            Dict with ``experiment`` config options
        bias_groups (list):
            List of 12 bias group configuration dictionaries.
        bands (list):
            List of 8 band configuration dictionaries.
    """
    def __init__(self):
        self.load_defaults()
        self.source_file = None

    def load_defaults(self):
        self.bands = [band_defaults.copy() for _ in range(8)]
        self.bias_groups = [bg_defaults.copy() for _ in range(12)]
        self.exp = exp_defaults.copy()

    @classmethod
    def from_dict(cls, data):
        """
        Creates a DeviceConfig object from a dictionary.
        """
        self = cls()
        for k, v in data['bands'].items():
            if k.lower().startswith('amc'):  # Key formatted like AMC[i]
                amc_index = int(k[4])
                self._load_amc(amc_index, v)
            if k.lower().startswith('band'):  # Key formatted like Band[i]
                band_index = int(k[5])
                self.bands[band_index].update(v)


        # Loads bias groups from config file
        for i, bg in enumerate(self.bias_groups):
            for k, vlist in data['bias_groups'].items():
                self.bias_groups[i][k] = vlist[i]

        # Loads experiment config data
        self.exp.update(data['experiment'])
        return self

    @classmethod
    def from_yaml(cls, filename):
        """
        Creates a DeviceConfig object from a dev-cfg yaml file.

        Args:
            filename (path): path to device-config file.
        """
        filename = os.path.abspath(os.path.expandvars(filename))
        if not os.path.exists(filename):
            # Just return default device cfg
            print(f"File {filename} doesn't exist! Creating it wth default params")
            self = cls()
            self.source_file = filename
            self.update_file()
            return self
        else:
            with open(filename) as f:
                data = yaml.safe_load(f)
            self = cls.from_dict(data)
            self.source_file = filename
            return self

    def _load_amc(self, amc_index, data):
        """Loads amc data all at once"""
        for k, v in data.items():
            if len(v) != 4:
                raise ValueError("All data in AMC entry must be formatted as a "
                                 "list of length 4.")
            for i in range(4):
                band_index = amc_index*4 + i
                self.bands[band_index][k] = v[i]

    def dump(self, path, clobber=False):
        """
        Dumps all device configuration info to a file.

        Args:
            path (path):
                Location to dump config info to. This file must not already
                exist.
            clobber (bool):
                If true will overwrite existing file. Defaults to false.
        """
        if os.path.exists(path) and not clobber:
            raise FileExistsError(f"Can't dump device config! Path {path}"
                                   "already exists!!")

        def _format_yml(val):
            """Converts np dtypes to python types for yaml files"""
            if hasattr(val, 'dtype'):
                return val.item()
            elif isinstance(val, tuple):
                return list(val)
            else:
                return val

        data = YamlReps.Odict()
        data['experiment'] = {
            k: _format_yml(v) for k, v in self.exp.items()
        }

        data['bias_groups'] = {
            k: YamlReps.FlowSeq([_format_yml(bg[k]) for bg in self.bias_groups])
            for k in self.bias_groups[0].keys()
        }
        data['bands'] = YamlReps.Odict([
            (f'AMC[{i}]', {
                k: YamlReps.FlowSeq([
                    _format_yml(b[k]) for b in self.bands[4*i:4*i+4]
                ])
                for k in self.bands[0].keys()
            }) for i in [0, 1]])

        with open(path, 'w') as f:
            yaml.dump(data, f)

    def update_band(self, band_index, data, update_file=False):
        """
        Updates band configuration object.

        Args:
            band_index (int 0-7):
                Index of band toupdate
            data (dict):
                Dictionary of parameters to update. All parameters must exist in
                the loaded dev-cfg file.
        """
        band = self.bands[band_index]
        for k, v in data.items():
            if k not in band.keys():
                for b in self.bands:
                    b[k] = None
            band[k] = v
        if update_file and self.source_file is not None:
            self.update_file()

    def update_bias_group(self, bg_index, data, update_file=False):
        """
        Updates bias group configuration object.

        Args:
            bg_index (int 0-11):
                Index of bias group to update.
            data (dict):
                Dictionary of parameters to update. All parameters must exist in
                the loaded dev-cfg file.
        """
        bg = self.bias_groups[bg_index]
        for k, v in data.items():
            if k not in bg.keys():
                for _bg in self.bias_groups:
                    _bg[k] = None
            bg[k] = v
        if update_file and self.source_file is not None:
            self.update_file()

    def update_experiment(self, data, update_file=False):
        """
        Updates ``experiment`` configuration object.

        Args:
            data (dict):
                Dictionary of parameters to update. All parameters must exist in
                the loaded dev-cfg file.
        """
        for k, v in data.items():
            self.exp[k] = v
        if update_file and self.source_file is not None:
            self.update_file()

    def update_file(self):
        """
        Updates the device cfg file.1
        """
        self.dump(self.source_file, clobber=True)


    def apply_to_pysmurf_instance(self, S, load_tune=True):
        """
        Applies basic device config params (amplifier biases, attens,
        tone_powers) to a pysmurf instance based on the device cfg values. Note
        that this does not set any of the tracking-related params since this
        shouldn't replace tracking_setup.
        """
        S.set_amplifier_bias(
            bias_hemt=self.exp['amp_hemt_Vg'],
            bias_50k=self.exp['amp_50k_Vg']
        )

        if load_tune:
            tunefile = self.exp['tunefile']
            if os.path.exists(tunefile):
                S.load_tune(tunefile)
            else:
                print(f"Cannot load tunefile {tunefile} because it doesn't exist")

        for b in S._bands:
            band_cfg = self.bands[b]
            if 'uc_att' in band_cfg:
                S.set_att_uc(b, band_cfg['uc_att'])
            if 'dc_att' in band_cfg:
                S.set_att_dc(b, band_cfg['dc_att'])
            if 'drive' in band_cfg:
                # Sets the tone power of all enabled channels to `drive`
                amp_scales = S.get_amplitude_scale_array(b)
                amp_scales[amp_scales != 0] = band_cfg['drive']
                S.set_amplitude_scale_array(b, amp_scales)


def make_parser(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser()

    parser.add_argument_group("DetConfig Options")
    parser.add_argument('--sys-file',
                        help="Path to sys-config file. "
                             "Defaults to ``$SMURF_CONFIG_DIR/sys_config.yml``")
    parser.add_argument('--dev-file',
                        help="Path to device-config file. "
                             "Defaults to the path specified in the sys_config.")
    parser.add_argument('--pysmurf-file',
                        help="Path to Pysmurf config file. "
                        "Defaults to the path specified in the sys_config.")
    parser.add_argument('--uxm-file',
                        help="Path to the uxm file. "
                        "Defaults to the path specified in the sys_config")
    parser.add_argument('--slot', '-N', type=int, help="Smurf slot")
    parser.add_argument('--dump-configs', '-D', action='store_true',
                        help="If true, all config info will be written to "
                             "the pysmurf output directory")
    return parser

class DetConfig:
    """
    General Configuration class for SODETLIB.

    Attributes:
        sys_file (path): Path to sys-config file used.
        dev_file (path): Path to device-config file used
        pysmurf_file (path): Path to pysmurf-config file used.
        uxm_file (path): Path to uxm file
        sys (dict):
            System configuration dictionary, generated from the sys_config.yml
            file or ``--sys-file`` command line argument.
        dev (dict):
            Device configuration dictionary, generated from the device_config
            file specified in the sys_config or specified by the ``--dev-file``
            command line argument
    """
    def __init__(self, slot=None, sys_file=None, dev_file=None,
                 pysmurf_file=None, uxm_file=None):
        self.sys_file = sys_file
        self.dev_file = dev_file
        self.pysmurf_file = pysmurf_file
        self.uxm_file = uxm_file
        self.slot = slot
        self.sys = None
        self.dev: DeviceConfig = None
        self.dump = None  # Whether config files should be dumped

        self.S = None

        self._parser: argparse.ArgumentParser = None
        self._argparse_args = None


    def parse_args(self, parser=None, args=None):
        """
        Parses command line arguments along with det_config arguments and
        loads the correct configuration. See ``load_config`` for how config
        files are determined.

        Args:
            parser (argparse.ArgumentParser, optional):
                custom argparse parser to parse args with. If not specified,
                will create its own.
            args (list, optional):
                List of command line arguments to parse.
                Defaults to the command line args.
        """
        self._argparse_args = args
        parser = make_parser(parser)
        self._parser = parser

        args = parser.parse_args(args=args)

        if args.sys_file is None:
            args.sys_file = self.sys_file
        if args.pysmurf_file is None:
            args.pysmurf_file = self.pysmurf_file
        if args.dev_file is None:
            args.dev_file = self.dev_file
        if args.uxm_file is None:
            args.uxm_file = self.uxm_file

        self.load_config_files(
            slot=args.slot, sys_file=args.sys_file, dev_file=args.dev_file,
            pysmurf_file=args.pysmurf_file, uxm_file=args.uxm_file
        )

        return args

    def load_config_files(self, slot=None, sys_file=None, dev_file=None,
                          pysmurf_file=None, uxm_file=None):
        """
        Loads configuration files. If arguments are not specified, sensible
        defaults will be chosen.

        Args:
            slot (int, optional):
                pysmurf slot number. If None and there is only a single slot in
                the sys_file, it will use that slot. Otherwise, it will throw
                and error.
            sys_file (path, optional):
                Path to sys config file. If None, defaults to
                $SMURF_CONFIG_DIR/sys_config.yml
            dev_file (path, optional):
                Path to the Device file. If None, defaults to the device file
                specified in the sys-config file (device_configs[slot-2]).
            pysmurf_file (path, optional):
                Path to pysmurf config file. If None, defaults to the file
                specified in the sys-config (pysmurf_configs[slot-2]).
            uxm_file (path, optional):
                Path to uxm file
        """
        # This should be the same for every smurf-srv
        if 'SMURF_CONFIG_DIR' in os.environ:
            cfg_dir = os.environ['SMURF_CONFIG_DIR']
        elif 'OCS_CONFIG_DIR' in os.environ:
            cfg_dir = os.environ['OCS_CONFIG_DIR']
        else:
            raise ValueError(
                "SMURF_CONFIG_DIR or OCS_CONFIG_DIR must be set in the environment"
            )

        if sys_file is None:
            self.sys_file = os.path.join(cfg_dir, 'sys_config.yml')
        else:
            self.sys_file = sys_file

        # Load system settings
        with open(self.sys_file) as f:
            self.sys = yaml.safe_load(f)

        if slot is not None:
            self.slot = slot
        elif len(self.sys['slot_order']) == 1:
            self.slot = self.sys['slot_order'][0]
        else:
            raise ValueError(
                "Slot could not be automatically determined from sys_config "
                "file! Must specify slot directly from command line with "
                "--slot argument. "
                f"Available slots are {self.sys['slot_order']}."
            )

        slot_cfg = self.sys['slots'][f'SLOT[{self.slot}]']
        self.stream_id = slot_cfg['stream_id']

        if dev_file is None:
            self.dev_file = os.path.abspath(os.path.expandvars(slot_cfg['device_config']))
        else:
            self.dev_file = os.path.abspath(os.path.expandvars(dev_file))
        self.dev = DeviceConfig.from_yaml(self.dev_file)

        # Gets the pysmurf config file
        if pysmurf_file is None:
            self.pysmurf_file = os.path.expandvars(slot_cfg['pysmurf_config'])
        else:
            self.pysmurf_file = pysmurf_file

        if uxm_file is None:
            if 'uxm_file' in slot_cfg:
                self.uxm_file = os.path.expandvars(slot_cfg['uxm_file'])
        else:
            self.uxm_file = uxm_file
        if self.uxm_file is not None:
            with open(self.uxm_file) as f:
                self.uxm = yaml.safe_load(f)
        else:
            self.uxm = None

    def dump_configs(self, output_dir=None, clobber=False, dump_rogue_tree=False):
        """
        Dumps any config information to an output directory

        Args:
            output_dir (path): Directory location to put config files.
            dump_rogue_truee : bool
                If True, will dump the pysmurf rogue tree to the configs dir.

        Returns:
            run_out (path): path to output run file
            sys_out (path): path to output sys file
            dev_out (path): path to output device file.
            uxm_out (path): path to uxm file
        """
        if output_dir is None:
            if not self.S:
                raise ValueError("output_dir cannot be None in offline mode")
            output_dir = os.path.join(
                    self.S.base_dir, self.S.date, self.stream_id, 
                    self.S.name, 'config', self.S.get_timestamp(),
                  )
        print(f"Dumping sodetlib configs to {output_dir}")
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)

        # Manually set publisher action because set_action decorator won't
        # work without S being the first function argument.
        if self.S:
            start_action = self.S.pub._action
            start_action_ts = self.S.pub._action_ts

        try:
            outfiles = {}
            if self.S:
                self.S.pub._action = "config"
                self.S.pub._action_ts = self.S.get_timestamp()

            # Dump run info
            run_info = YamlReps.Odict([
                ('sys_file', self.sys_file),
                ('dev_file', self.dev_file),
                ('pysmurf_file', self.pysmurf_file),
                ('slot', self.slot),
                ('cwd', os.getcwd()),
                ('argv',  ' '.join(sys.argv)),
                ('time', int(time.time())),
            ])
            if self._argparse_args is not None:
                run_info['cfg_args'] = self._argparse_args
            run_out = os.path.abspath(os.path.join(output_dir, 'run_info.yml'))
            outfiles['run'] = run_out
            with open(run_out, 'w') as f:
                yaml.dump(run_info, f)
            if self.S:
                self.S.pub.register_file(run_out, 'config', format='yaml')

            # Dump sys file
            sys_out = os.path.abspath(os.path.join(output_dir,
                                                   'sys_config.yml'))
            outfiles['sys'] = sys_out
            with open(sys_out, 'w') as f:
                yaml.dump(self.sys, f)
            if self.S:
                self.S.pub.register_file(sys_out, 'config', format='yaml')

            # Dump device file
            dev_out = os.path.abspath(os.path.join(output_dir, 'dev_cfg.yml'))
            outfiles['dev'] = dev_out
            self.dev.dump(dev_out, clobber=clobber)
            if self.S:
                self.S.pub.register_file(dev_out, 'config', format='yaml')

            # Copy pysmurf file
            pysmurf_out = os.path.join(output_dir, 'pysmurf_cfg.yml')
            pysmurf_out = os.path.abspath(pysmurf_out)
            outfiles['pysmurf'] = pysmurf_out
            shutil.copy(self.pysmurf_file, pysmurf_out)
            if self.S:
                self.S.pub.register_file(pysmurf_out, 'config', format='yaml')

            # Dump uxm file
            if self.uxm is not None:
                uxm_out = os.path.abspath(os.path.join(output_dir, 'uxm.yml'))
                outfiles['uxm'] = uxm_out
                with open(uxm_out, 'w') as f:
                    yaml.dump(self.uxm, f)
                if self.S:
                    self.S.pub.register_file(uxm_out, 'config', format='yaml')

            if dump_rogue_tree:
                print("Dumping state")
                state_file = os.path.abspath(os.path.join(output_dir,
                                                          'rogue_state.yaml'))
                outfiles.append(state_file)
                self.S.save_state(state_file)
        finally:
            # Restore publisher action so it doesn't mess up ongoing actions
            if self.S:
                self.S.pub._action = start_action
                self.S.pub._action_ts = start_action_ts

        return outfiles

    def get_smurf_control(self, offline=False, epics_root=None,
                          smurfpub_id=None, make_logfile=False, setup=False,
                          dump_configs=None, config_dir=None,
                          apply_dev_configs=False, load_device_tune=True,
                          **pysmurf_kwargs):
        """
        Creates pysmurf instance based off of configuration parameters.
        If not specified as keyword arguments ``epics_root`` and ``smurf_pub``
        will be created based on the slot and crate id's.

        Args:
            offline (bool):
                Whether to start pysmurf in offline mode. Defaults to False
            epics_root (str, optional):
                Pysmurf epics root. If none, it will be set to
                ``smurf_server_s<slot>``.
            smurfpub_id (str, optional):
                Pysmurf publisher ID. If None, will default to
                crate<crate_id>_slot<slot>.
            make_logfile (bool):
                Whether pysmurf should write logs to a file (True) or
                stdout(False). Defaults to stdout.
            setup (bool):
                Whether pysmurf should run a full setup. Defaults to False.
            dump_configs (bool):
                Whether all configuration settings should be dumped to pysmurf
                data directory. Defaults to True.
            **pysmurf_kwargs:
                Any additional arguments to be passed to pysmurf initialization.

        """
        import pysmurf.client

        slot_cfg = self.sys['slots'][f'SLOT[{self.slot}]']
        if epics_root is None:
            epics_root = f'smurf_server_s{self.slot}'
        if smurfpub_id is None:
            smurfpub_id = self.stream_id
        if dump_configs is None:
            dump_configs = self.dump

        # Pysmurf publisher will check this to determine publisher id.
        os.environ['SMURFPUB_ID'] = smurfpub_id

        if offline:
            S = pysmurf.client.SmurfControl(offline=True)
        else:
            S = pysmurf.client.SmurfControl(
                epics_root=epics_root, cfg_file=self.pysmurf_file, setup=setup,
                make_logfile=make_logfile, data_path_id=smurfpub_id,
                **pysmurf_kwargs)
        self.S = S
        # Lets just stash this in pysmurf...
        S._sodetlib_cfg = self

        # Dump config outputs
        if dump_configs:
            if config_dir is None:
                if not offline:
                    config_dir = os.path.join(
                        S.base_dir, S.date, smurfpub_id, S.name, 'config',
                        S.get_timestamp(),
                    )

                else:
                    config_dir = os.path.join('config', S.get_timestamp())
                    print("Warning! Being run in offline mode with no config "
                          f"directory specified! Writing to {config_dir}")
            self.dump_configs(config_dir)

        if apply_dev_configs:
            print("Applying device cfg parameters...")
            self.dev.apply_to_pysmurf_instance(S, load_tune=load_device_tune)

        return S
